#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ALAN API 
Last mod.,
Thu Aug 25 19:19:30 EDT 2022
"""
import codecs,sys,os
os.environ['TZ']='America/New_York' # set New_York EST as default
from optparse import OptionParser
import numpy as np
import pandas as pd
import pandas_datareader.data as web
import datetime
import re
import json
import os,cgi,cgitb
from _alan_calc import get_start_end,subDict,subDF,renameDict,getKeyVal
from _alan_calc import  sqlQuery,conn2mgdb,conn2pgdb,pqint,saferun
from _alan_str import find_mdb,dtCvt
from _alan_date import next_date,s2dt
from glob import glob,iglob
import base64
import subprocess
from bson.objectid import ObjectId

import pprint
#%matplotlib inline 
#import matplotlib.pyplot as plt 
from hashlib import md5
if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')

# set pandas global display formula
pd.set_option('display.max_colwidth', -1)
pd.options.display.float_format='{:,.2f}'.format

# set mongoDB database connection as globals()
mgDB=conn2mgdb(dbname='ara')
pgDB=conn2pgdb(dbname='ara')
currDate=datetime.datetime.now()

def verify_apikey(ak):
	if ak=='beyondbond_internal_testing_only':
		return 1
	ht=pd.read_csv("/apps/fafa/webDir/loginDir/.htpasswd.dat",sep="\t")
	aLst=list(ht['api_key'])
	return 1 if ak.lower() in aLst else 0

class myPrettyPrinter(pprint.PrettyPrinter):
	def format(self, object, context, maxlevels, level):
		if isinstance(object, unicode):
			#return (object.encode('utf8'), True, False)
			return (object, True, False)
		return pprint.PrettyPrinter.format(self, object, context, maxlevels, level)

def dft_tablename(search,instrument,tablename):
	if tablename is not None :
		return tablename
	if search == "search_list" :
		tablename='(select m.*,s.exch from spdr_list s, mapping_ticker_cik m where m.ticker=s.ticker and act_code=1 order by sector DESC,ticker) as m' if instrument=='stock' else 'mapping_series_label'
	elif search in ["search_history", "search_hist" ]:  
		tablename='prc_hist' if instrument=='stock' else 'ara_globalmacro_hist'
	#elif search == "search_quote" :
	#	tablename='yh_quote_curr' if instrument=='stock' else 'macro_hist_fred'
	#elif search == "search_comment" :
	#	tablename='ohlc_daily_comment_cn' if instrument=='stock' else 'ara_globalmacro_forecast_comment_cn'
	elif search == "search_factor" :
		tablename='ara_outlook_factor_temp' if instrument=='stock' else 'ara_globalmacro_forecast_comment_cn'
	else :
		tablename = None
	return tablename
	
def fmtSetting(x,pcsn=2,comma=','):
	from numpy import int64,int32
	if isinstance(x,float):
		return "{:{}.{}f}".format(x,comma,pcsn)
	elif isinstance(x,(int,int64,int32)):
		return "{:{}}".format(x,comma)
	else:
		return "{}".format(x)

def fmt0f(x):
	return fmtSetting(x,pcsn=0,comma='')

def fmt0a(x):
	return fmtSetting(x,pcsn=0,comma=',')

def fmt2a(x):
	return fmtSetting(x,pcsn=2,comma=',')

@saferun
def data_output(data,output='json',**optx):
	''' Convert dataFrame 'data' into type:'output' output
	'''
	if data is None or len(data)<1:
		ret=getKeyVal(optx,'ret',{})
		return ret
	elif '**ERROR' in data[:9] or not isinstance(data,pd.DataFrame):
		if  isinstance(data,(dict,list)):
			return json.dumps(data,default=dtCvt) #ensure_ascii=False
		return data
	elif output == 'dict' :
		orient=getKeyVal(optx,'orient','records')
		return data.to_dict(orient=orient)
	elif output == 'json' :
		orient=getKeyVal(optx,'orient','records')
		if sys.version_info.major == 2:
			return data.to_json(orient=orient,force_ascii=False).encode('utf8')
		return data.to_json(orient=orient,default_handler=dtCvt)
	elif output == 'json_by_column' :
		return data.to_json()
		#return json.loads(data.to_json())
		#return data.to_json(force_ascii=False).encode('utf8')
	elif output == 'html' : 
		float_fmt=getKeyVal(optx,'float_fmt','{:,.2f}')
		pd.options.display.float_format=float_fmt.format
		data = data.replace([np.nan, np.inf, -np.inf], '', regex=True)
		#return data.to_html(index=False)
		cfm = { 'pbdate':fmt0f,
			'closeDate':fmt0f,
			'epochs':fmt0f,
			'close':fmt2a,
			'marketCap':fmt0a,
			'dollarValue':fmt0a,
			'volume':fmt0a,
			'count':fmt0a}
		cfm=getKeyVal(optx,'cfm',cfm)
		return data.to_html(index=False,formatters=cfm)
	elif output == 'csv_by_comma' :
		index=getKeyVal(optx,'index',False)
		return data.to_csv(sep=",",index=index)
	elif output == 'tsv' :
		index=getKeyVal(optx,'index',False)
		return data.to_csv(sep="\t",index=index)
	elif output == 'csv' :
		sep=getKeyVal(optx,'sep','|')
		index=getKeyVal(optx,'index',False)
		return data.to_csv(sep=sep,index=index)
	elif output in ['DF','df','DataFrame','dataframe'] :
		return data
	else:
		sep=getKeyVal(optx,'sep','|')
		index=getKeyVal(optx,'index',False)
		return data.to_csv(sep=sep,index=index)

def xql_factor_comment(ticker,fdLst,factor='technical',lang='',subtopic='attr'):
	xlang = "_{}".format(lang) if lang=='cn' else ''
	if factor =='overall':
		if subtopic=='attr':
			tablename="ara_outlook_factor_temp"
		else:
			tablename="ara_outlook_factor_recommend{}".format(xlang)
	else:
		if subtopic=='attr':
			tablename="ara_outlook_attr_{}".format(factor)
		else:
			tablename="ara_outlook_attr_{}_comment{}".format(factor,xlang)
	
	if ticker=='*':
		xqr="SELECT {} FROM {} WHERE factor='{}'".format(fdLst,tablename,factor)
	else:
		xqr="SELECT {} FROM {} WHERE ticker='{}' AND factor='{}'".format(fdLst,tablename,ticker,factor)
	return xqr

def geteach_comment(ticker,fdLst,tablename=None,lang=None,dbname="ara",hostname='localhost',topic='factor',subtopic='',factor='',**optx):
	if tablename is None:
		return None
	colx='ticker'
	if factor in ['technical','valuation','quality','macro','overall']:
		xqr=xql_factor_comment(ticker,fdLst,factor=factor,lang=lang,subtopic=subtopic)
	elif ticker=='*':
		xqr="SELECT {} FROM {}".format(fdLst,tablename)
	else:
		xqr="SELECT {} FROM {} WHERE {}='{}' ".format(fdLst,tablename,colx,ticker)
	pqint( xqr ,file=sys.stderr)
	try:
		data=pd.read_sql(xqr,con=pgDB)
		if ticker=='*':
			xqr="SELECT ticker,ranking,ntotal from ara_ranking_list where category='AI' and subtype='SP500'".format(ticker)
		else:
			xqr="SELECT ticker,ranking,ntotal from ara_ranking_list where category='AI' and subtype='SP500' and ticker='{}'".format(ticker)
		rankX=pd.read_sql(xqr,con=pgDB)
		data = data.merge(rankX,on='ticker')
		#pqint(data.tail(2) ,file=sys.stderr)
	except Exception as e:
		pqint(str(e), 'failed to pull pricing data' ,file=sys.stderr)
		return None
	return data


def geteach_peers(tkLst,fdLst,subtopic='',output='csv',tablename=None,lang=None,dbname="ara",hostname='localhost',**optx):
	from iex_peers import iex_peers
	df,peerLst,peerInfo=iex_peers(tkLst)
	#if subtopic!='detailed':
	#	return peerLst
	if fdLst !='*':	
		field = list(set(fdLst.split(',')) & set(df.columns))
		
		df = df[field]
	data = df # data = data_output(df,output)
	pqint( df.tail(2) ,file=sys.stderr)
	return data

def peers_comment(tkLst,fdLst,**optx):
	return geteach_peers(tkLst,fdLst,**optx)

def rerun_dss(ticker=''):
	'''
	rerun daily_single_stock via forking subprocess
	'''
	data=''
	if len(ticker)<1:
		return json.dumps(dict(err="no ticker assigned",err_code=400))
	tkLst=sqlQuery("SELECT ticker from mapping_ticker_cik where ticker='{}'".format(ticker))
	if len(tkLst)<1:
		return json.dumps(dict(err="ticker not available",err_code=402))
	try:
		xcmd="cd /apps/fafa/pyx/alan/; python3 _alan_mp4.py daily_single_stock {} 1".format(ticker)
		sys.stderr.write("==RERUN: {}\n".format(xcmd))
		myproc = subprocess.Popen(xcmd,shell=True)
		if myproc.poll() != None:
			sys.stderr.write("**WARNING: {} not running\n".format(xcmd))
		else:
			sys.stderr.write("=={}: {} still running\n".format(ticker,xcmd))
		return json.dumps(dict(err="Analysis will be available shortly.",err_code=304))
	except Exception as e:
		sys.stderr.write("**WARNING: {} on {}\n".format(str(e),xcmd))
		return json.dumps(dict(err=str(e),err_code=403))

def report_comment(tkLst,fdLst,lang='cn',dbname='ara',hostname='localhost',limit=1,**optx):
	outTF=optx.pop('outTF',True)
	tablename='daily_single_stock'
	field=optx.get('field',{})
	if tkLst is not None and len(tkLst)>0:
		ticker=tkLst[0].upper()
		jobj={"ticker":ticker}
	else:
		ticker=''
		jobj={}
	zsrt={'pbdt'}
	data=[]
	try:
		#- for debugging as directly call mongodb
		#zsrt=[(x,-1) for x in zsrt]
		#host="{}:27017".format(hostname)
		#mgDB=MongoClient(host)[dbname]
		#data=list(mgDB[tablename].find(jobj,{'_id':0},sort=zsrt).limit(limit) )
		data,_,err=find_mdb(jobj,clientM=None,tablename=tablename,dbname=dbname,sortLst=zsrt,limit=1,dfTF=False)
		#if len(xtmp)>0:
		#	data = [{x:y for x,y in xtmp[0].items() if not isinstance(y,(dict))}]
	except Exception as e:
		sys.stderr.write("**ERROR: {},{} @{}\n".format(ticker,str(e),'report_comment'))
		data=[]

	if len(data)<1:
		ret=rerun_dss(ticker)
		sys.stderr.write("**WARNING: {} @{}\n".format(ret,'report_comment'))
		return ret

	# data = data_output(data,output)
	pqint( "===report_comment():\nFind:{},Field:{},Sort:{}\n".format(jobj,field,zsrt) ,file=sys.stderr)
	pqint( " --from tablename: {}".format(tablename) ,file=sys.stderr)
	pqint( " --DF[top]:\n{}".format(data) ,file=sys.stderr)
	return data

def news1_comment(tkLst,fdLst,**optx):
	from rssnews import rssnews
	outTF=optx.pop('outTF',True)
	opts={}
	for x,y in optx.items():
		if y is not None and x not in ['topic','dbname']:
			opts.update({x:y})
	if len(tkLst)>0:
		opts.update(tklst=tkLst)
	opts.update(outTF=False,nth=":")
	sys.stderr.write(" --OPTS:{} @{}\n".format(opts,"news_comment"))
	data=rssnews(**opts)
	if outTF and not isinstance(data,pd.DataFrame):
		data = pd.DataFrame(data)
	sys.stderr.write(" --data:{}\n".format(data)[:100]+"\n")
	return data

def pubdates_ob(pubdates='',key="pubDate"):
	from pandas import Timestamp as ts
	if not pubdates:
		return {}
	x1, x2 = pubdates.split(',')
	if all([x1,x2]):
		ob = {key:{"$gte":ts(x1),"$lte":ts(x2)}}
	elif  x1:
		ob = {key:{"$gte":ts(x1)}}
	elif  x2:
		ob = {key:{"$lte":ts(x2)}}
	else:
		ob = {}
	return ob

def get_news_query(queryObj,**opts):
	if 'source' in opts and opts['source']: # for multi-source selection
		queryObj.update({'source':{'$in':opts['source'].split(',')}})
	if 'url' in opts and opts['url']:
		queryObj.update({'url':opts['url']})
	if 'newstopic' in opts and opts['newstopic']:
		queryObj.update({'topic':opts['newstopic']})
	if 'title' in opts and opts['title']:
		queryObj.update({'title':{'$regex':opts['title'],'$options':'i'}})
	if 'title_cn' in opts and opts['title_cn']:
		queryObj.update({'title_cn':{'$regex':opts['title_cn'],'$options':'i'}})
	if 'summary' in opts and opts['summary']:
		queryObj.update({'summary':{'$regex':opts['summary'],'$options':'i'}})
	if 'pubdates' in opts and opts['pubdates']:
		ob = pubdates_ob(opts['pubdates'])
		queryObj.update(ob)
	return queryObj

def news_comment(tkLst,fdLst,lang='cn',dbname='ara',hostname='localhost',output='html',limit=200,**optx):
	print("===@news_comment tkLst:{}, optx: {}".format(tkLst,optx),file=sys.stderr)
	from _alan_str import ssh2mg
	serverM,clientM=ssh2mg()
	sys.stderr.write(" --serverM:{},clientM:{}\n".format(serverM,clientM))
	from mongo_en2cn import mongo_en2cn as me
	outTF=optx.pop('outTF',True)
	tablename='rssNews'
	jobj={}
	subtopic=optx.get('subtopic','company')
	sortLst=['pubDate']
	data=[]
	if subtopic=='company' and len(tkLst)>0:
		ticker=tkLst[0]
		from rssCompany import run_rssCompany as rrc
		data = rrc(ticker)
		print("===Run Company News:{},\n --{}".format(ticker,data)[:100],file=sys.stderr)
	elif subtopic=='translate':
		link=optx.pop('link','')
		_id=optx.pop('_id','')
		src=optx.pop('fromLang','en')
		dest=optx.pop('toLang','zh-tw')
		tablename=optx.pop('tablenews','rssCompany')
		if link:
			jobj={'link':link}
		elif not _id:
			return []
		else:
			jobj={'_id':ObjectId(_id)}
		field={'title','title_cn','summary','summary_cn'}
		print("===Run Translation:",jobj,tablename,field,src,dest,file=sys.stderr)
		data = me(jobj,clientM=clientM,tablename=tablename,field=field,src=src,dest=dest)
	else:
		jobj = get_news_query(jobj,**optx)
		pqint(" --->news query jobj:\n",jobj,file=sys.stderr)

		field={'_id','ticker','title','title_cn','description','summary','summary_cn','pubDate','link','source'}
		data,_,_=find_mdb(jobj,clientM=clientM,tablename=tablename,dbname=dbname,field=field,sortLst=sortLst,limit=limit,dfTF=outTF)
		if isinstance(data,pd.DataFrame):
			data.drop_duplicates(subset =["pubDate","link"], keep = 'first', inplace = True)
	if fdLst not in ['*',''] and fdLst is not None:
		fields=fdLst.split(',')
		data = subDict(data,fields)
	if outTF and not isinstance(data,pd.DataFrame):
		data = pd.DataFrame(data)
	if serverM:
		clientM.close()
		serverM.stop()
	return data

def hourly_comment(tkLst,fdLst,tablename=None,lang=None,dbname="ara",hostname='localhost',output=None,start=None,end=None,topic=None,subtopic=None,**optx):
	from _alan_str import find_mdb
	if fdLst=='*':
		field = {}
	else:
		field = set(fdLst.split(','))
	limit=0
	instrument = getKeyVal(optx,'instrument','stock')
	rpt_hm = getKeyVal(optx,'rpt_hm',None)
	hhmm = getKeyVal(optx,'hhmm',None)
	pqint(tablename,optx,file=sys.stderr)
	#if subtopic.lower() == 'sector':
	tablename="mkt_briefing_media" # use if gtts-cli is working for audio/video
	#tablename="mkt_briefing_details" # otherwise
	if subtopic == 'sector':
		tablename="hourly_report"
		if len(tkLst)<1:
			jobj={"ticker":{"$in":["^GSPC","^DJI","^IXIC"]}}
		sortLst=['rpt_time']
		field={'ticker','comment','rpt_time','rpt_hm','rpt_status','title','label'}
		limit=3
	elif rpt_hm is not None or len(tkLst)>0:
		tablename="mkt_briefing_details"
		jobj={}
		if rpt_hm is not None:
			hmLst = [int(x) for x in rpt_hm.split(',')]
			jobj.update(rpt_hm={"$in":hmLst})
		if len(tkLst)>0 and tkLst[0]!='$' :
			jobj.update(ticker={"$in":tkLst})
		if len(field)<1:
			field = {"ticker","label","rpt_hm","rpt_time","cprice","xprice","comment","rpt_date","pngname"}
		sortLst=['rpt_time']
	elif hhmm is not None:
		hmLst = [int(x) for x in hhmm.split(',')]
		jobj={"hhmm":{"$in":hmLst}}
		sortLst=['pbdt']
	else:
		jobj={}
		sortLst=['pbdt']
		limit=1
	outTF = getKeyVal(optx,'outTF',True)
	df,_,_=find_mdb(jobj,tablename=tablename,dbname=dbname,field=field,sortLst=sortLst,limit=limit,dfTF=outTF)
	data = df # data = data_output(df,output)
	pqint( "===hourly_comment():\nFind:{},Field:{},Sort:{}\n".format(jobj,field,sortLst) ,file=sys.stderr)
	pqint( " --tkLst: {},fdLst: {}".format(tkLst,fdLst) ,file=sys.stderr)
	pqint( " --from tablename: {}".format(tablename) ,file=sys.stderr)
	pqint( " --DF keys:\n{}".format(df[0].keys())+"\n" ,file=sys.stderr)
	pqint( " --DF:\n{}".format(df)[:200]+"\n" ,file=sys.stderr)
	return data

def theme_comment(tkLst,fdLst,subtopic='ipo',**opts):
	from _alan_rmc import run_topic_theme
	outTF=opts.pop('outTF',True)
	if subtopic.lower() in ['majorplayer','media','ipo','etffund']:
		funcName= "topic_{}_{}".format("theme",subtopic)
		sys.stderr.write("===RUNNING {} with OPTS:{}\n".format(funcName,opts))
		df=run_topic_theme(funcName=funcName,subtopic=subtopic,outTF=outTF,**opts)
	if len(df)<0:
		return {}
	if fdLst not in ['*',''] and fdLst is not None:
		fields=fdLst.split(',')
		df = subDict(df,fields)
	if len(df)<0:
		return []
	if isinstance(df,pd.DataFrame):
		sys.stderr.write("==OPTS:{},FIELDS:{}, DATA:\n{}".format(opts,fdLst,df.tail(2)))
	data = df # data = data_output(df,**opts)
	return data

def industry_comment(tkLst,fdLst,output=None,**opts):
	df = getlist_filter(fdLst=fdLst,subtopic='industry')
	if not (tkLst is None or len(tkLst)<1 or tkLst[0] in ['','*']):
		df = df[df['ticker'].isin(tkLst)]
	if len(df)<1:
		return None
	data = df # data = data_output(df,output)
	return data

def search_comment(tkLst,fdLst,**opts):
	topicLst='hourly|news|report|theme|peers|industry|MFRM|news1'.split('|')
	topic=getKeyVal(opts,'topic','MFRM')
	if topic not in topicLst:
		return None
	argName="{}_comment".format(topic)
	if topic in topicLst and argName in globals():
		pqint("==RUNNING {}() Inputs:{}".format(argName,opts),file=sys.stderr)
		try:
			data=globals()[argName](tkLst,fdLst,**opts)
		except Exception as e:
			pqint("**ERROR:{} to run {}".format(str(e),argName),file=sys.stderr)
			data={}
		return data
	output=opts.pop('output',None)
	data=pd.DataFrame()
	optx=subDict(opts,['tablename','lang','dbname','hostname','topic','subtopic','factor'])
	if tkLst[0]=='*':
		data=geteach_comment('*',fdLst,**optx)
		# data = data_output(data,output)
		return data
	for ticker in tkLst:
		ret=geteach_comment(ticker,fdLst,**optx)
		if ret is not None and len(ret)>0:
			data=data.append(ret,ignore_index=True)
		else:
			continue
	# data = data_output(data,output)
	if len(data)<1:
		return None
	return data

def search_factor(tkLst,fdLst,**opts):
	opts.update(subtopic='attr')
	return search_comment(tkLst,fdLst,**opts)

def geteach_quote(ticker,fdLst,tablename=None,lang=None,dbname="ara",hostname='localhost',colx='ticker'):
	if tablename is None:
		return None
	xtmp="select {} from {} WHERE {}='{}' order by pbdate DESC limit 2"
	xqr=xtmp.format(fdLst,tablename,colx,ticker)
	sys.stderr.write(" --geteach_quote from pgDB::{}\n".format(tablename))
	sys.stderr.write(" --SQL:\n{}\n".format(xqr))
	try:
		data=pd.read_sql(xqr,con=pgDB)
		data=renameDict(data,{"series":"ticker","value":"close"})
		colv = 'close'
		if data.shape[0]>1 and 'change' not in data.columns:
			chg = data[colv].iloc[0]-data[colv].iloc[1]
			pchg = data[colv].iloc[0]/data[colv].iloc[1]-1
			data = data.iloc[0]
			data['change']=chg
			data['changePercent']=pchg
		sys.stderr.write(" --DATA[{}]:\n{}\n".format(len(data),data))
	except Exception as e:
		sys.stderr.write("**ERROR:{} @ {}\n".format(str(e),'geteach_quote'))
		return None
	return data

def getlist_quotes(tkLst,fdLst,tablename="iex_sector_quote",lang=None,dbname="ara",hostname='localhost',colx='symbol'):
	from _alan_str import find_mdb
	dd = []
	pqint( tkLst ,file=sys.stderr)
	for ticker in tkLst:
		ret,_,_=find_mdb({colx:ticker},tablename=tablename,dbname=dbname)
		pqint( ret,ticker,tablename,dbname ,file=sys.stderr)
		dd = dd + ret
	df = pd.DataFrame(dd)
	if fdLst != '*':
		field = set(fdLst.split(','))
		newcol = list(field & set(df.columns))
		df = df[newcol]
	return df

def geteach_financials_history(ticker,fdLst,**opts):
	debugTF = getKeyVal(opts,'debugTF',False)
	subtopic = getKeyVal(opts,'subtopic','')
	src = getKeyVal(opts,'src','zacks')
	jobj={'ticker':ticker}	
	if subtopic == 'eps': # and src=='zacks':
		tablename='earnings_zacks'
		jd=list(mgDB[tablename].find(jobj,{'_id':0},sort=[("pbdate",-1)]))
		if len(jd)<1:
			return []
		datax = pd.DataFrame(jd)
		datax = renameDict(datax,{'pChg':'epsSuprise%','ESP':'epsSuprise'})
	elif subtopic == 'eps' and src=='yh': # No longer used
		tablename='earnings_yh'
		jobj.update({'actualEPS':{'$ne':np.nan}})
		jd=list(mgDB[tablename].find(jobj,{'_id':0},sort=[("pbdate",-1)]))
		if len(jd)<1:
			return []
		for j,xd in enumerate(jd):
			jd[j]['EPSReportDate']="{}{}{}{}-{}{}-{}{}".format(*list(str(jd[j]['pbdate'])))
		datax = pd.DataFrame(jd)
	elif subtopic == 'roe':
		freq = getKeyVal(opts,'freq','Q')
		tablename='qS_IS_{}'.format('A' if freq=='A' else 'Q')
		jd=list(mgDB[tablename].find(jobj,{'_id':0}))
		if len(jd)<1:
			return []
		datax = pd.DataFrame(jd) if len(jd)>0 else []

		tablename='qS_BS_{}'.format('A' if freq=='A' else 'Q')
		jd=list(mgDB[tablename].find(jobj,{'_id':0}))
		if len(jd)>0:
			for j,xd in enumerate(jd):
				a=jd[j]['pbdate']
				jd[j]['reportDate']='{:04d}-{:02d}-{:02d}'.format(int(a/10000),int(a/100)%100,a%100)
				jd[j]['freq']=freq
			d2 = pd.DataFrame(jd)
			datax = datax.merge(d2,on=['ticker','pbdate','endDate'],how='inner')
			if debugTF:
				sys.stderr.write("=====geteach_financials_history() d2:{}".format(d2))
		if 'netIncome' in datax.columns and 'totalStockholderEquity' in datax.columns:
			datax['roe'] = datax['netIncome']/datax['totalStockholderEquity']
		if debugTF:
			sys.stderr.write("=====geteach_financials_history() datax:{}".format(datax))
		datax=subDF(datax,['ticker','roe','pbdate','reportDate','freq','endDate','netIncome','totalStockholderEquity'])
		sys.stderr.write(" ---datax:{}\n{}\n".format(ticker,datax))
	else:
		return []
	return datax

def geteach_financials_historyOLD(ticker,fdLst,**opts):
	colx='ticker'
	if 'subtopic' in opts and opts['subtopic'] == 'eps':
		tablename='earnings_hist_yh'
		xqTmp = "select {} from {} where {}='{}' order by pbdate"
		xqr=xqTmp.format(fdLst,tablename,colx,ticker)
	elif 'subtopic' in opts and opts['subtopic'] == 'roe':
		freq = opts['freq'] if 'freq' in opts else 'Q'
		tablename='iex_financials_hist'
		xqTmp = "select (\"netIncome\"::float/\"shareholderEquity\"*100.) as roe, {} from {} where {}='{}' and freq='{}' order by pbdate"
		xqr=xqTmp.format(fdLst,tablename,colx,ticker,freq)
	sys.stderr.write(" --{} SQL:\n{}\n".format('geteach_financials_history',xqr))
	datax=pd.read_sql(xqr,con=pgDB)
	return datax

def geteach_minute_web(ticker,fdLst,**opts):
	''' pull minute data from web 
	'''
	from _alan_calc import get_minute_iex 
	# get_minute_yh DEPRECATED, yh_spark_hist is used
	from yh_chart import yh_spark_hist as ysh, runOTF;
	src = opts.pop('src','iex')
	src= "yh" if '^' in ticker or '=' in ticker else src
	sys.stderr.write("---Get minute data {} from WEB::{}\n".format(ticker,src.upper()))
	if src=='yh':
		#datax = get_minute_yh(ticker,ranged='1d',tsTF=True,debugTF=False)
		#d=ysh([ticker],saveDB=False,range='1d',types='chart',interval='5m',debugTF=True)
		d=runOTF(ysh,ticker,deltaTolerance=900,types='chart',tablename='yh_chart_hist',zpk=['ticker','pbdt'],range='15m',interval='5m',debugTF=True,dbname='ara')
		if len(d)>0:
			datax=pd.DataFrame(d)
		else:
			return []
	else:
		datax = get_minute_iex(ticker,ranged='1d',date=None,tsTF=True,debugTF=False)
	return datax

def geteach_minute_db(ticker,fdLst,**opts):
	''' pull minute data from database 
	'''
	subtopic=getKeyVal(opts,'subtopic','')
	src = opts.pop('src','iex')
	src= "yh" if '^' in ticker or '=' in ticker else src
	tablename="{}_spark_hist".format(src)
	sys.stderr.write("---Get minute data {} from MDB::{}\n".format(ticker,tablename))
	try:
		jobj={"ticker":ticker}
		# Get latest pbdate
		xd = mgDB[tablename].find_one({"$query":jobj,"$orderby":{"epochs":-1}},{"pbdate":1,"_id":0})
		jobj.update(xd)
		if subtopic in ['sparkline','spark']:
			d = mgDB[tablename].find(jobj,{"_id":0},sort=[("epochs",-1)]).limit(20)
		else:
			d = mgDB[tablename].find(jobj,{"_id":0})
		if d is None:
			return {}
		data=list(d)
		if len(data)<1:
			sys.stderr.write(" --No data found from {}\n".format(tablename))
			return {}
		datax = pd.DataFrame(data)
		#datax,_,_ = find_mdb({"ticker":ticker},clientM=mgDB,tablename=tablename,sortLst={"epochs"},dfTF=True)
	except Exception as e:
		sys.stderr.write("**ERROR:{} @ {}\n".format(str(e),'geteach_minute_db'))
		datax={}
	return datax

def geteach_minute_history(ticker,fdLst,**opts):
	''' pull minute data from database (default)
	OR web if 'webTF' is True
	'''
	webTF=subtopic=getKeyVal(opts,'webTF',False)
	if webTF:
		opts.update(src='yh')
		return geteach_minute_web(ticker,fdLst,**opts)
	data=geteach_minute_db(ticker,fdLst,**opts)
	if data is None or len(data)<1:
		sys.stderr.write("**WARNING:{}\n".format("data not found in DB, live pulling"))
		opts.update(src='yh')
		return geteach_minute_web(ticker,fdLst,**opts)
	return data

def geteach_daily_history(ticker,fdLst,**opts):
	from _alan_calc import pullStockHistory
	from yh_chart import yh_spark_hist as ysh, runOTF
	datax = []
	try:
		datax = pullStockHistory(ticker,pgDB=pgDB,**opts)
		if datax is None or len(datax)<1:
			sys.stderr.write("**WARNING:{}\n".format("data not found in DB, live pulling"))
			datax=runOTF(ysh,ticker,deltaTolerance=86400,types='chart',tablename='yh_daily_hist',zpk=['ticker','pbdt'],range='1y',interval='1d',debugTF=True,dbname='ara')
			datax=pd.DataFrame(datax)
	except Exception as e:
		pqint("**ERROR:{} @ {}, opts:\n{}".format(str(e),'geteach_daily_history',opts) ,file=sys.stderr)
	return datax

def geteach_history(ticker,fdLst,**opts):
	# for topic list 'financial','daily','minute'
	topic = getKeyVal(opts,'topic','daily') 
	argName = "geteach_{}_history".format(topic)
	try:
		if argName not in globals():
			argName="geteach_daily_history"
		sys.stderr.write(" RUNNING Ticker:{}[fdLst:{}]:topic:{} of {}()\n".format(ticker,fdLst,topic,argName) )
		argFunc = globals()[argName]
		df = argFunc(ticker,fdLst,**opts)
	except Exception as e:
		sys.stderr.write("**ERROR: {} @{}()".format(str(e),argName) )
		df = ''
	return df

def get_sector_etfname(tkLst,**opts):
	if 'topic' in opts and 'subtopic' in opts:
		if opts['topic']=='market' and opts['subtopic']=='sector':
			tkLst = list(pd.read_sql("select etfname from spdr_sector",con=pgDB)['etfname'].values)
	return tkLst

def get_valid_DB_quote(ticker='AAPL',tablename="yh_quote_curr",deltaTolerance=86400):
	jobj={"ticker":ticker}
	ret = list(mgDB[tablename].find(jobj,{"_id":0},sort=[("epochs",-1)]).limit(1))
	#ret,_,_=find_mdb(jobj,tablename=tablename,dbname="ara")
	if len(ret)<1:
		return []
	pbdtMod=ret[0]['pbdt']
	pbdtCurr=pd.Timestamp(pd.Timestamp.now())
	deltaPassed=pd.Timedelta(pbdtCurr - pbdtMod).total_seconds()
	if deltaPassed>deltaTolerance:
		return []
	pqint( "==DB[{}] quote {}:\n{}".format(tablename,ticker,ret) ,file=sys.stderr)
	return ret

def search_quote(tkLst,fdLst,**opts):
	from yh_chart import yh_spark_hist as ysh;
	tkLst=get_sector_etfname(tkLst,**opts)
	instrument = getKeyVal(opts,'instrument','stock')
	outTF = getKeyVal(opts,'outTF',True)
	hostname,dbname,tablename,lang = getKeyVal(opts,['hostname','dbname','tablename','lang'],['localhost','ara',None,None])
	colx='ticker' if instrument=='stock' else 'series'
	data=[]
	opts.pop('ticker',None)
	for ticker in tkLst:
		sys.stderr.write("---Get {} @ search_quote\n".format(ticker))
		try:
			# get quotes from MDB::"yh_quote_curr" for yahoo source indices setup in the PGDB::'mapping_series_label'
			if instrument=='stock' or re.search(r'[=^.]',ticker):
				mktLst =['^GSPC','^DJI','^IXIC','^SOX']
				if ticker.upper() in mktLst:
					tablename="market_indicator_quote"
				else:
					tablename="yh_quote_curr"
				ret=get_valid_DB_quote(ticker,tablename=tablename)
			else: # get quotes for all fields from pgDB
				ret=geteach_quote(ticker,fdLst='*',tablename=tablename,lang=lang,dbname=dbname,hostname=hostname,colx=colx)
			if instrument=='stock' and len(ret)<1: # get live quote
				pqint( "==Get live quote:{} @ {}".format(ticker,'yh_spark_hist()') ,file=sys.stderr)
				ret = ysh(ticker,types='quote',dfTF=False,**opts)
				pqint( "==Live quote {}:\n{}".format(ticker,ret) ,file=sys.stderr)
			if ret is not None and len(ret)>0:
				#data=data.append(ret,ignore_index=True)
				data.extend(ret)
			else:
				continue
		except Exception as e:
			pqint( "**ERROR:{} @ {}".format(str(e),search_quote) ,file=sys.stderr)
			continue
	if len(data)<1:
		return None
	data=pd.DataFrame(data)
	data=renameDict(data,{'pchg':'changePercent','xclose':'prevClose'})
	colx=['ticker','epochs','open','high','low','close','volume','xclose','change','pchg','pbdt','hhmm','pbdate','changePercent','prevClose','marketCap']
	if fdLst is not None:
		if fdLst.lower()=='all':
			return data
		elif fdLst.lower()=='':
			pass
		else:
			fieldLst = list(set(fdLst.split(',')) & set(data.columns))
			colx = fieldLst if len(fieldLst)>0 else colx
	data=subDF(data,colx)
	#if outTF:
	#	output = opts.pop('output','json')
	#	data = data_output(data,output)
	return data

def search_hist(tkLst,fdLst,**opts):
	output=getKeyVal(opts,'output','json')
	topic=getKeyVal(opts,'topic','daily')
	opts.pop('ticker',None)
	data=pd.DataFrame()
	dd=[]
	for ticker in tkLst:
		df=geteach_history(ticker,fdLst,**opts)
		sys.stderr.write(" --DF: {}\n{}\n".format(ticker,type(df)))
		if isinstance(df,pd.DataFrame) and len(df)>0:
			data=data.append(df,ignore_index=True)
		elif isinstance(df,list) and len(df)>0:
			dd=dd.extend(df)
		else:
			continue
	if len(dd)>0:
		data=pd.DataFrame(dd)
	sys.stderr.write(" --DATA[{}] tail:\n{}\n".format(len(data),data.tail()))
	if topic not in ['daily','minute']:
		return data # data_output(data,output)
	data=renameDict(data,{"name":"ticker"})
	colx=['ticker','epochs','open','high','low','close','volume','xclose','change','pchg','pbdt','hhmm','pbdate']
	data=subDF(data,colx)
	return data # data_output(data,output)

def search_history(tkLst,fdLst='',**opts):
	return search_hist(tkLst,fdLst=fdLst,**opts)

# DEPRECATED
def getlist_filterOLD(fdLst='',subtopic='mostvalue',start=None,id=None):
	from _alan_str import stock_screener as bbpre
	scrDct={'mostactive':'most_actives', 'mostvalue':'most_values',
		'gainers':'day_gainers', 'losers':'day_losers', 'industry':'sector_gainers',
		'5%':'5%','-5%':'-5%','+5%':'+5%',
		'day_gainers':'day_gainers', 'day_losers':'day_losers', 'sector_gainers':'sector_gainers',
		'sector_losers':'sector_losers', 'most_actives':'most_actives', 'most_values':'most_values'}
	s=subtopic.lower()
	sID=scrDct.pop(s,'most_values')
	df=bbpre(sID)
	if 'close' not in df and 'price' in df:
		df['close']=df['price']
	return df

def getlist_filter(fdLst='',subtopic='mostvalue',start=None,id=None):
	from _alan_str import stock_screener as bbpre
	if subtopic.lower()=='mostactive': # 爆量
		df=bbpre('most_actives',addiFilter=1)
	elif subtopic.lower()=='mostvalue': # 金額爆量
		df=bbpre('most_values',addiFilter=1)
	elif subtopic.lower()=='gainers': # 牛方
		df=bbpre('day_gainers',addiFilter='abs(changePercent)>1&price>4.99')
	elif subtopic.lower()=='losers': # 熊方
		df=bbpre('day_losers',addiFilter='abs(changePercent)>1&price>4.99')
	elif subtopic in ['5%','+5%']:
		df=bbpre('day_gainers',addiFilter=2)
	elif subtopic == '-5%':
		df=bbpre('day_losers',addiFilter=2)
	elif subtopic.lower() == 'industry':
		df=bbpre('sector_gainers')
	else: # most_active as default
		df=bbpre('most_actives',addiFilter=1)
	if 'close' not in df and 'price' in df:
		df['close']=df['price']
	if 'ranking' not in df:
		df['ranking'] = [int(x)+1 for x in df.index.values]
	return df

def getlist_market(fdLst,subtopic='sector',start=None,id=None):
	if subtopic=='theme':
		xqr = "SELECT theme FROM theme_list group by theme"
		if id is not None:
			xqTmp = "SELECT {} FROM theme_list where theme='{}'"
			xqr = xqTmp.format(fdLst,id)
	else:
		if id is not None:
			xqTmp = "SELECT {} FROM mapping_ticker_cik WHERE sector='{}'"
			xqr = xqTmp.format(fdLst,id)
		else:
			xqTmp = "SELECT {} FROM spdr_sector WHERE sector not like '%%Index'"
			xqr = xqTmp.format(fdLst)
	sys.stderr.write(" --SQL:{}\n".format(xqr))
	data=pd.read_sql(xqr,con=pgDB)
	return data

def getlist_recommend(fdLst,subtopic='AI',start=None,id=None):
	return getlist_AI(fdLst,subtopic=subtopic,start=start,id=id)

def getlist_AI(fdLst,subtopic='AI',start=None,id=None):
	"""
	subtopic list [ai|industry|value|growth]
	"""
	subtopic = subtopic.lower()
	if start is None:
		start="2018-10-24"
	if fdLst == '*':
		fieldLst = fdLst
	else:
		fieldLst = '"{}"'.format('","'.join(fdLst.split(',')))
	
	if subtopic=='growth': # for z=0.225 for 60% uptrend
		xqTmp = """SELECT {} FROM ara_outlook_factor_temp WHERE sector not like '%%Index' and factor='overall' and zscore>=0.255 ORDER BY zscore DESC limit 10"""
		xqr = xqTmp.format(fieldLst)
	elif subtopic=='value':
		xqTmp = """SELECT {} FROM ara_outlook_factor_temp WHERE sector not like '%%Index' and factor='valuation' and zscore>=0.255 ORDER BY zscore DESC limit 10"""
		xqr = xqTmp.format(fieldLst)
	elif subtopic=='watchlist':
		xqTmp = """SELECT {} FROM ara_outlook_factor_temp WHERE sector not like '%%Index' and factor='valuation' and zscore<=-0.0 ORDER BY zscore DESC limit 10"""
		xqr = xqTmp.format(fieldLst)
	elif subtopic=='alert':
		xqTmp = """SELECT {} FROM ara_outlook_factor_temp WHERE sector not like '%%Index' and factor='valuation' and zscore<=-0.255 ORDER BY zscore limit 10"""
		xqr = xqTmp.format(fieldLst)
	elif subtopic=='industry':
		xqTmp = """SELECT {} FROM ara_outlook_factor_recommend_cn WHERE length(recommend)>0 and sector like '%%Index' ORDER BY confidence DESC"""
		xqr = xqTmp.format(fieldLst)
	elif subtopic=='alert':
		xqTmp = """SELECT {} FROM ara_outlook_factor_recommend_cn WHERE length(recommend)>0 and sector not like '%%Index' ORDER BY confidence DESC limit 10"""
		xqr = xqTmp.format(fieldLst)
	else:
		return None

	pqint(xqr ,file=sys.stderr)
	data=pd.read_sql(xqr,con=pgDB)
	#data['ranking'] = [int(x)+1 for x in data.index.values]
	return data

import getpass
def geteach_list(tkLst,fieldLst,tablename=None,lang=None,dbname="ara",hostname='localhost'):
	if tablename is None:
		return None
	if fieldLst is None:
		fieldLst='*'
	xqr="SELECT {1} FROM {0}".format(tablename,fieldLst)
	if tablename == 'mapping_series_label':
		xqr = xqr + " WHERE mkt_list=1 ORDER BY category_seq,category_label_seq"
	elif tkLst[0]!='*':
		inLst=json.dumps(tkLst)[1:-1].replace('"','\'')
		xqr = xqr + " WHERE ticker in ({}) ORDER BY ticker".format(inLst)
	pqint( xqr ,file=sys.stderr)
	try:
		data=pd.read_sql(xqr,con=pgDB)
		pqint(data.tail(2) ,file=sys.stderr)
	except Exception as e:
		pqint("***ERROR: {} @ {}()".format(str(e),"geteach_list") ,file=sys.stderr)
		return "***ERROR: {} @ {}()".format(str(e),"geteach_list")
		#return None
	return data

def search_list(tkLst,fdLst,**opts):
	topic,start,subtopic,output,id = getKeyVal(opts,['topic','start','subtopic','output','id'],[None,None,None,None,None])
	hostname,dbname,tablename,lang = getKeyVal(opts,['hostname','dbname','tablename','lang'],['localhost','ara',None,None])
	pqint("==>{}::topic:{},subtopic:{},field:{}".format('search_list',topic,subtopic,fdLst) ,file=sys.stderr)
	if hasattr(subtopic,'__len__') is not True:
		subtopic = ''
	if topic in ['filter','recommend','AI','market']: # select either [getlist_filter|getlist_recommend]
		funcName = "getlist_{}".format(topic)
		if funcName in globals():
			searchListFunc = globals()[funcName]
			pqint( "Applying {}: {}".format(funcName,searchListFunc) ,file=sys.stderr)
			data = searchListFunc(fdLst,subtopic=subtopic,start=start,id=id)
			# data = data_output(df,output)
			return data
	elif topic in ['ytdRtn','sp500','cdr','ipo']:
		#from ytdRtn_calc import ytdRtn_calc,ytdOTF
		from ytdRtn_cdr import ytdRtn_calc
		if subtopic is None or len(subtopic)<1: 
			subtopic='sector'
		if start is None or len(start)<1: 
			start=20210101
		pqint("---{}::topic:{},subtopic:{},start:{}".format('ytdRtn_calc',topic,subtopic,start) ,file=sys.stderr)
		data = ytdRtn_calc(topic=topic,start=start,group=subtopic)[0]
		return data
		#tablename='ytdRtn_{}'.format(subtopic)
		#data=ytdOTF(ytdRtn_calc,start=start,deltaTolerance=900,tablename=tablename,zpkChk=['group'],zpk=[subtopic],group=subtopic,debugTF=True,dbname='ara')
		#sys.stderr.write(" --ytdRtn_calc data:{}\n{}\n".format(type(data),data))
		#outTF=opts.pop('outTF',True)
		#if outTF:
		#	data=pd.DataFrame(data)
		#return data
	elif topic in ['peers']:
			data = geteach_peers(tkLst,fdLst,subtopic=subtopic,output=output)
			return data
		
	if tablename is None:
		tablename='spdr_list'
	data=geteach_list(tkLst,fdLst,tablename=tablename,lang=lang,dbname=dbname,hostname=hostname)
	# data = data_output(data,output)
	return data

def search_allocation(tkLst,fdLst,**opts):
	output = getKeyVal(opts,'output',None)
	from bl_portopt2 import bl_example
	try:
		data = bl_example(tkLst)
		data['mktCapWeigh'] = data['mktCapWeigh']*100.0
		data['weight'] = data['weight']*100.0
		data['rrt'] = data['rrt']*100.0
		pd.options.display.float_format = '{:.2f}'.format
		# data = data_output(data,output)
	except Exception as e:
		data = None
	return data

def search_mp3(tkLst,fdLst,**opts):
	tagContent='''
	<audio class="audioClass" controls>
	  <source src="{}" type="audio/mpeg">
	  Your browser does not support the audio tag.
	</audio>
	'''
	attrType=getKeyVal(opts,'attrType','src') # tag|src|content for tag_setup|src_filename|content
	tagContent=getKeyVal(opts,'tagContent',tagContent)
	limit=getKeyVal(opts,'limit',30) 
	aliasDir=getKeyVal(opts,'aliasDir','/apps/fafa/pyx/alan/US/mp3_hourly')
	routeDir=getKeyVal(opts,'routeDir','/US/mp3_hourly')
	filename=getKeyVal(opts,'filename','*.mp3')
	fpath="{}/{}".format(aliasDir,filename)
	fpLst = sorted(iglob(fpath), key=os.path.getmtime, reverse=True)
	sys.stderr.write("===Search {}\n".format(attrType))
	sys.stderr.write(" --searching:{}, found:{}\n".format(fpath,fpLst))
	data=[]
	for fname in fpLst[:limit]:
		fp=fname.replace(aliasDir,routeDir)
		if attrType=='tag':
			data=tagContent.format(fp)
			return data
		elif attrType=='content':
			fp=open(fname,'rb')
			data=fp.read()
			fp.close()
			return data
		else:
			data=fp
			return data
			#data.append(fp)
	#sys.stderr.write(" --Ret Data:\n{}\n".format(data))
	return data

def search_mp4(tkLst,fdLst,**opts):
	tagContent='''
	<video class="videoClass" controls>
	  <source src="{}" type="video/mp4">
	  Your browser does not support the video tag.
	</video>
	'''
	opts.pop('tagContent',None)
	return search_mp3(tkLst,fdLst,tagContent=tagContent,**opts)

def search_img(tkLst,fdLst,**opts):
	tagContent='''<img class="imgClass" src="{}" >'''
	opts.pop('tagContent',None)
	return search_mp3(tkLst,fdLst,tagContent=tagContent,**opts)

def search_txt(tkLst,fdLst,**opts):
	attrType=getKeyVal(opts,'attrType','src') # tag|src|content for tag_setup|src_filename|content
	opts.pop('attrType',None)
	data= search_mp3(tkLst,fdLst,attrType=attrType,**opts)
	return data

def cgi_api(optx={'pprint':'N','api_key':''}):
	import os 
	curdir = os.path.dirname(os.path.realpath(__file__))
	helpfile =  "{}/{}".format(curdir,'alanapi_help.html')
	xf, _ = opt_alanapi()
	xf.update(optx)
	cgitb.enable(display=0,logdir="/apps/fafa/https/html/log/",format="text")
	cssStyle='''<style>)
	table {border-collapse: collapse;font-size:11pt}
	thead {background:lightgray;font-size:14pt;}
	table tr {border:1px solid black}
	tbody tr:nth-child(even){background:lightgray;}
	</style>
	'''
	mf = cgi.FieldStorage()
	sys.stderr.write("===== START CGI arguments:{}:{}\n{}\n".format(s2dt(),list(mf),mf))
	for ky in mf:
		va=mf[ky].value
		if ky[-2:] == 'TF':
			va = True if va[:1].lower() in ['t','y','1'] else False
		xf[ky]=va
	search,output,attrType=getKeyVal(xf,['search','output','attrType'],['list',None,'src'])
	### No action and show help
	if xf['search'] is None:	
		print("Content-type:text/html;charset=utf-8\r\n\r\n")
		print(open(helpfile).read())
		return None

	### No api_key match
	if verify_apikey(xf['api_key']) < 1:
		print("Content-type:text/html;charset=utf-8\r\n\r\n")
		print("<H2 style='color:red;'>{}</H2>".format("203: Non-Authoritative Information due to invalid api_key"))
		print(open(helpfile).read())
		return None

	### Start setup cgi parameters xf
	if 'ticker' in xf: 
		symLst=xf['ticker'].split(',')
	else:
		symLst=None
	### Run model
	try:
		fdLst = xf['field']
		xf.pop('field',None)
		sys.stderr.write("===START run_alanapi {}\n".format(s2dt()))
		sys.stderr.write(" --tkLst: {}\n".format(symLst))
		sys.stderr.write(" --fdLst: {}\n".format(fdLst))
		sys.stderr.write(" --OPTS: {}\n".format(xf))
		sys.stderr.write(" --OUT OPTS:{}:{}:{}\n".format(search,output,attrType))
		data=run_alanapi(symLst,fdLst,**xf)
	except Exception as e:
		sys.stderr.write("**ERROR: {} @ {}()\n".format(str(e),"cgi_api"))
		sys.stderr.write(" --OPTS: {}\n".format(xf))
	### START CGI OUTPUT 
	if data is None or (type(data) == 'str' and len(data.strip())<1):
		print("Content-type:text/html;charset=utf-8\r\n\r\n")
		print("<H2 style='color:red;'>{}</H2>".format("204: No Content"))
		print(open(helpfile).read())
	elif output=='json':
		print("Content-type:application/json;charset=utf-8\r\n\r\n")
		print(data)
	elif output=='dict':
		sys.stdout.write("Content-type:text/plain;charset=utf-8\r\n\r\n")
		pprint.pprint(data)
	elif search=='mp3' and attrType=='content': # streaming data
		try:
			data=base64.b64encode(data).decode()
			data="data:audio/mpeg;base64,"+data
		except Exception as e:
			sys.stderr.write("**ERROR mp3:{}\n{}\n".format(str(e),type(data)))
		#OR
		print("Content-type:text/html;\r\n\r\n")
		data='<embed class=audioClass src="'+data+'" hidden="false" autostart="false" loop="false"/>'
		print(data)
		sys.stderr.write("{}\n".format(data[:100]))
	elif search=='mp4' and attrType=='content': # streaming data
		#print("Content-type:video/mp4;charset=base64\r\n\r\n")
		data=base64.b64encode(data).decode()
		data="data:video/mp4;base64,"+data
		print("Content-type:text/html;\r\n\r\n")
		data='<embed class=videoClass src="'+data+'" hidden="false" autostart="false" loop="false"/>'
		print(data)
	elif search=='img' and attrType=='content': # streaming data
		#print("Content-type:image/svg+xml;\r\n\r\n")
		data=base64.b64encode(data).decode()
		data="data:image/svg+xml;base64,"+data
		print("Content-type:text/html;\r\n\r\n")
		data='<embed class=imgClass src="'+data+'" hidden="false" loop="false"/>'
		print(data)
	elif output=='html':
		print("Content-type:text/html;charset=utf-8\r\n\r\n")
		print(cssStyle)
		tblStyle=getKeyVal(xf,'tblStyle',cssStyle)
		print(data)
	else:
		if output in ['csv','tsv']:
			sys.stdout.write("Content-type:text/plain;charset=utf-8\r\n\r\n")
		else:
			print("Content-type:text/html;charset=utf-8\r\n\r\n")
			pageHeader=getKeyVal(xf,'pageHeader','ALAN API Page')
			if pageHeader is not None:
				print("<header><title>{}</title></header>\r\n".format(pageHeader))
		print(data)
	sys.stderr.write("===END cgi_api @ {}\n".format(s2dt()))
	return data

from pymongo import MongoClient
from bson import json_util
import json
def search_ohlc(tkLst,fdLst,tablename=None,lang=None,dbname="ara",hostname='localhost',output=None,start=None,end=None,topic=None,subtopic=None):
	host="localhost:27017";dbname="ara";tablename="ohlc_pattern"
	dbM=MongoClient(host)[dbname]
	name="morning_star"
	data=[]
	for ticker in tkLst:
		#pData=dbM[tablename].find({"ticker":ticker,"name":name})
		if ticker=='*':
			pData=dbM[tablename].find()
		else:
			pData=dbM[tablename].find({"ticker":ticker})
		mobj= json.loads(json_util.dumps(pData))
		data.append(mobj)
	return data

def run_alanapi(symLst,fdLst,**opts):
	'''
	list of keys:
	['lang', 'subtopic', 'search', 'end', 'start', 'debugTF', 'hostname', 'tablename', 'topic', 'field', 'instrument', 'output', 'dbname']
	'''
	search, instrument, hostname, dbname, tablename = opts['search'],opts['instrument'],opts['hostname'],opts['dbname'],opts['tablename']
	if search == None:
		search='search_list'
	elif search[:7] != 'search_':
		search = 'search_{}'.format(search)
	if symLst is None or len(symLst)<1:
		symLst=['*'] if search[-5:] == '_list' else []
	try:
		opts['tablename']=dft_tablename(search,instrument,tablename)
		pqint(" --- {} {}, instrument:{}, tablename:{}\n".format(search,symLst,instrument,tablename) ,file=sys.stderr)
		searchFunc=globals()[search]
		opts.pop('search',None)
		data=searchFunc(symLst,fdLst,**opts)
	except Exception as e:
		pqint('**ERROR: failed to run {}: {}'.format(searchFunc,str(e)) ,file=sys.stderr)
		return None
	outTF = opts.pop('outTF',True)
	if outTF:
		output = opts.pop('output','json')
		data = data_output(data,output)
	return data
	

def opt_alanapi():
	parser = OptionParser(usage="usage: %prog [option] SYMBOL ...", version="%prog 0.7",
		description="API for ALAN")
	parser.add_option("-s","--start",action="store",dest="start",
		help="start YYYY-MM-DD (default: 3-year-ago)")
	parser.add_option("-e","--end",action="store",dest="end",
		help="end YYYY-MM-DD (default: today)")
	parser.add_option("","--src",action="store",dest="src",default="iex",
		help="source [fred|yh|iex](default: iex)")
	parser.add_option("","--host",action="store",dest="hostname",default="localhost",
		help="db host (default: localhost)")
	parser.add_option("","--database",action="store",dest="dbname",default="ara",
		help="database (default: eSTAR_2)")
	parser.add_option("-t","--table",action="store",dest="tablename",
		help="db tablename (default: None)")
	parser.add_option("","--field",action="store",dest="field",default="*",
		help="database column field (default: *)")
	parser.add_option("","--instrument",action="store",dest="instrument",default="stock",
		help="instrument type [stock|macro] (default: None)")
	parser.add_option("-o","--output",action="store",dest="output",
		help="output type [json|csv|html] (default: None)")
	parser.add_option("-l","--lang",action="store",dest="lang",default="en",
		help="language (default: en)")
	parser.add_option("","--search",action="store",dest="search",
		help="SEARCH action (default: None)")
	parser.add_option("","--topic",action="store",dest="topic",
		help="TOPIC to search  (default: None)")
	parser.add_option("","--subtopic",action="store",dest="subtopic",
		help="SUBTOPIC (default: None)")
	parser.add_option("","--extra_js",action="store",dest="extraJS",
		help="extra JSON in DICT format.")
	parser.add_option("","--extra_xs",action="store",dest="extraXS",
		help="extra excutable string in k1=v1;k2=v2; format")
	parser.add_option("","--debug",action="store_true",dest="debugTF",default=False,
		help="debugging (default: False)")
	(options, args) = parser.parse_args()
	try:
		opts = vars(options)
		from _alan_str import extra_opts
		extra_opts(opts,xkey='extraJS',method='JS',updTF=True)
		extra_opts(opts,xkey='extraXS',method='XS',updTF=True)
		opts.pop('extraXS',None)
		opts.pop('extraJS',None)
	except Exception as e:
		sys.stderr.write("**ERROR:{} @{}\n".format(str(e),'opt_csv2plot'))
	return opts, args

def main():
	""" API for ALAN
		Usage: alanapi.py [option] SYMBOL ...
	"""
	pp = myPrettyPrinter(indent=4,width=20)
	xreq=os.getenv('REQUEST_METHOD')
	data=''
	if xreq in ('GET','POST') : # WEB MODE
			data=cgi_api()
	else:
		(opts, args)=opt_alanapi()
		pqint( (opts, args) ,file=sys.stderr)
		symLst = args if len(args)>0 else None
		fdLst = opts['field']
		opts.pop('field',None)
		sys.stderr.write("===START: {}:{}:{}\n".format(symLst,fdLst,opts))
		data=run_alanapi(symLst,fdLst,**opts)
		print(data)
	return data

if __name__ == '__main__':
	try:
		data = main()
	except Exception as e:
		sys.stderr.write("**ERROR:{} @ {}\n".format(str(e),'main'))



#!/usr/bin/env python3
''' Live datafeed from yahoo quotes
Update tables:
mDB: ara::yh_spark_hist, ara::yh_quote_curr
pgDB: ara::yh_quote_curr (isEoD=True)

Usage of:
+ From commandline: 
python3 yh_chart.py IBM,AAPL
python3 yh_chart.py IBM AAPL
echo "IBM AAPL" | python3 yh_chart.py -

#-------------------------------------------------------------------------
+ For a list of tickers 'tkLst' with direct function call for pricing data
  -- To run Intraday for yh_spark_hist (CRON every 15 minutes 930-1600 M-F)
python3 -c "from yh_chart import yh_spark_hist as ysh;ysh(ranged='15m',interval='5m',debugTF=True)"
  -- To run EoD yh_quote_curr and yh_spark_hist (CRON at 1600 M-F)
python3 -c "from yh_chart import yh_spark_hist as ysh;ysh(types='quote',isEoD=True,debugTF=True)"
  -- To run SoD yh_quote_curr
python3 -c "from yh_chart import yh_spark_hist as ysh;ysh(types='quote',debugTF=True)"
  -- To run intraday yh_quote_market
printf "^GSPC ^IXIC ^DJI ^SOX" | python3 -c "import sys;from yh_chart import yh_spark_hist as ysh;ysh(sys.stdin.read().split(),t2='market_indicator_quote',types='quote',debugTF=True)"

#-------------------------------------------------------------------------
+ For specific 'ticker' of either non-existed ticker or if not updated in the last 'deltaTolerance' seconds
  -- To setup onTheFly stock quote for non-existed ticker or if not updated in the last 'deltaTolerance' seconds
python3 -c "ticker='AAPL';from yh_chart import yh_quote_comparison as yqc, runOTF;ret=runOTF(yqc,ticker,deltaTolerance=3600,tablename='yh_quote_curr',zpk=['ticker']);"
  -- To setup onTheFly stock intra-day spark history for non-existed ticker or if not updated in the last 'deltaTolerance' seconds
python3 -c "tkLst='AAPL';from yh_chart import yh_spark_hist as ysh, runOTF;ret=runOTF(ysh,ticker,deltaTolerance=900,tablename='yh_spark_hist',zpk=['ticker','pbdt'],ranged='15m',interval='5m',debugTF=True,dbname='test');"
  -- To setup onTheFly financials 'modules' for non-existed ticker or if not updated in the last 'deltaTolerance' seconds (86400 for 1-day)
python3 -c "ticker='AAPL';from yh_chart import runOTF, yh_financials as yh;ret=runOTF(yh,ticker,deltaTolerance=86400,modules='incomeStatementHistoryQuarterly',tablename='yh_financials',dbname='yh',zpkChk=['ticker','module'],zpk=['ticker','module','pbdate']);"
  -- To setup onTheFly stock intra-day chart history for non-existed ticker or if not updated in the last 'deltaTolerance' seconds (TBD)
python3 -c "tkLst='AAPL';from yh_chart import yh_spark_hist as ysh, runOTF;ret=runOTF(ysh,ticker,deltaTolerance=900,types='chart',tablename='yh_chart_hist',zpk=['ticker','pbdt'],ranged='15m',interval='5m',debugTF=True,dbname='ara');"

#-------------------------------------------------------------------------
+ From Direct function call for financial keyStatistics
python3 -c "from yh_chart import qS_keyStatistics as yks;yks(saveDB=True,debugTF=True)"
  -- To run Quarterly income_statement
python3 -c "from yh_chart import yh_financials_batch as yf;yf(modules='incomeStatementHistoryQuarterly',tablename='yh_IS_Q',zpk={'ticker','pbdate'});"
python3 -c "from yh_chart import yh_financials_batch as yf;yf(modules='incomeStatementHistoryQuarterly',tablename='yh_IS_Q',zpk={'ticker','pbdate'},useDB=True);"
  -- To run Annual income_statement
python3 -c "from yh_chart import yh_financials_batch as yf;yf(modules='incomeStatementHistory',tablename='yh_IS_A',zpk={'ticker','pbdate'});"
+ NOTE:
  -- below 3 calls are equivalent (1st one perferred, see onTheFlyDB_tst for better reference)
python3 -c "ticker='DAL';from yh_chart import runOTF, yh_financials as yh;ret=runOTF(yh,ticker,deltaTolerance=86400,modules='incomeStatementHistoryQuarterly',tablename='yh_financials',dbname='yh',zpkChk=['ticker','module'],zpk=['ticker','module','pbdate']);"
python3 -c "tkLst=['DAL'];from yh_chart import func2mdb as yhb;dd=yhb(tkLst,dbname='test',tablename='test1',funcN='yh_batchTypes',zpk={'ticker','pbdate'},types='quoteSummary',modules='incomeStatementHistoryQuarterly',debugTF=True);print(dd)"
python3 -c "tkLst=['DAL'];from yh_chart import yh_financials_batch as yf;yf(tkLst,modules='incomeStatementHistory',dbname='test',tablename='test1',zpk={'ticker','module','pbdate'});"
python3 -c "ticker=['AAPL','DAL'];from yh_chart import runOTF, yh_financials as yf;ret=runOTF(yf,ticker,deltaTolerance=86400*90,modules='"summaryProfile"',tablename='"yh_summaryProfile"',dbname='ara',zpkChk=['ticker','module'],zpk=['ticker','module','pbdate']);"

  -- To run summaryProfile  and save to MDB: yh_summaryProfile table
python3 -c "tkLst=['SHAK'];from yh_chart import func2mdb as yhb;dd=yhb(tkLst,tablename='yh_summaryProfile',funcN='yh_batchTypes',zpk=['ticker'],types='quoteSummary',modules='summaryProfile',debugTF=True);print(dd)"

  -- To run summaryProfile and show output
python3 -c "tkLst=['SHAK'];from yh_chart import yh_batchTypes as yhb;dd=yhb(tkLst,types='quoteSummary',modules='summaryProfile',debugTF=True);print(dd)"
  -- To run defaultKeyStatistics and show output
python3 -c "tkLst=['AAPL','IBM'];from yh_chart import yh_batchTypes as yhb;dd=yhb(tkLst,types='quoteSummary',modules='defaultKeyStatistics',debugTF=True);print(dd)"
  -- To add quote of tickers
python3 -c "from yh_chart import func2mdb as yhq; yhq(['XLRE'],tablename='yh_quote_curr',dbname='ara',funcN='yh_quote_comparison',zpk={'ticker'});"

  -- DEPRECATED
python3 -c "from yh_chart import yh_quote_curr as yqc;d=yqc()"
python3 -c "from yh_chart import func2mdb as yhq; yhq(['XLRE'],tablename='yh_quote_curr',dbname='ara',funcN='yh_quote_comparison',zpk={'ticker'});"
python3 -c "from yh_chart import yh_hist_query as yhq;d=yhq(['AAPL'],types='spark');print(d)"
python3 -c "from yh_chart import yh_hist_query as yhq;d=yhq(['AAPL'],types='quote');print(d)"
python3 -c "from yh_chart import yh_quote_comparison as yqc;print(yqc(['AAPL']))"

Last mod.,
Wed May 27 10:20:54 EDT 2020
-----------------------------------------------------------------------------------------------------
'''
import sys, datetime
import requests
import numpy as np
import pandas as pd
from pandas.io.json import json_normalize
from _alan_str import write2mdb,find_mdb,insert_mdb,upsert_mdb,get_arg2func
from _alan_calc import getKeyVal,conn2pgdb,conn2mgdb,renameDict,subDict,subDF,saferun,safeRunArg
headers={'Content-Type': 'text/html', 'Accept': 'application/json', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}

def epoch_parser(x,s=1000): return datetime.datetime.fromtimestamp(int(x/s))

def pull_act_tickers(sql=None,engine=None):
	if not sql:
		sql="select ticker from mapping_ticker_cik where act_code=1 ORDER BY ticker"
	try:
		from _alan_calc import sqlQuery
		df = sqlQuery(sql,engine=engine)
	except Exception as e:
		df={}
	return df

def list2chunk(v,n=100):
	''' partition an array 'v' into arrays limit to 'n' elements
	'''
	import numpy as np
	return [v[i:i+max(1,n)] for i in np.arange(0, len(v), n)]


def yh_spark_hist(tkLst=None,**optx):
	sys.stderr.write("++yh_spark_hist():optx: {}\n".format(optx))
	dd = yh_batchSpark(tkLst=tkLst,**optx)
	if optx.pop('dfTF',False) and not isinstance(dd,pd.DataFrame):
		dd = pd.DataFrame(dd)
	if optx.pop('tsTF',False) and isinstance(dd,pd.DataFrame) and \
		not isinstance(dd.index,(datetime.date, datetime.datetime)):
		if 'pbdt' in dd.columns:
			dd.set_index('pbdt',drop=True,inplace=True)
		elif 'epochs' in dd.columns:
			pbdt = np.vectorize(epoch_parser)(dd['epochs'])
			dd.set_index(pd.DatetimeIndex(pbdt),inplace=True)
		dd.index.name='date'
	return dd

def useWeb(jobj={'ticker':'^GSPC'},colx='pbdt',dbname='ara',tablename='yh_quote_curr',mmGap=30,**optx):
	'''return [True|Flase] action for using web or DB
	based on 'tablename' 'colx' field

	'''
	from _alan_calc import conn2mgdb
	debugTF=getKeyVal(optx,'debugTF',False)
	webTF=optx.pop('webTF',None)
	if webTF is not None and isinstance(webTF,bool):
		return webTF
	cdt=getKeyVal(optx,'cdt',datetime.datetime.now())
	mgDB=conn2mgdb(dbname=dbname)

	dc = mgDB[tablename].find_one({"$query":jobj,"$orderby":{colx:-1}},{colx:1,"_id":0})
	if len(dc)<1:
		return True
	pdt=dc[colx]
	mmPassed=pd.Timedelta(cdt - pdt).total_seconds() / 60.0
	webTF = mmPassed>mmGap
	if debugTF is True:
		sys.stderr.write("{}|{}|{}|{}|{}\n".format("webTF","cdt","pdt","mmPassed","mmGap"))
		sys.stderr.write("{}|{}|{}|{:.0f}|{}\n".format(webTF,cdt,pdt,mmPassed,mmGap))
	return webTF

def yh_quote_curr(tkLst=None,screenerTF=False,dbname='ara',tablename='yh_quote_curr',zpk={'ticker'},m2pTF=False,webTF=None,**optx):
	'''
	create yh_quote_curr table
	'''
	if tkLst is None or len(tkLst)<1:
		from _alan_calc import sqlQuery
		tkDF=sqlQuery('select ticker from mapping_ticker_cik where act_code=1')
		tkLst=list(tkDF['ticker'])
	try:
		jobj=dict(ticker=tkLst[0])
		if webTF is None:
			webTF = useWeb(jobj=jobj,dbname=dbname,tablename=tablename,**optx)
		if not webTF:
			jobj = {'ticker':{'$in':tkLst}}
			d,_,_ = find_mdb(jobj,dbname=dbname,tablename=tablename,dfTF=True)
			sys.stderr.write("===Using data from {}::{}\n".format(dbname,tablename))
			return d
		d=yh_quote_comparison(tkLst,screenerTF=screenerTF,dbname=dbname,tablename=tablename,zpk=zpk,**optx)
	except Exception as e:
		sys.stderr.write("**ERROR: {}:{}:{}\n".format("yh_quote_curr()","@ yh_quote_comparison",str(e)))
		d=[]
	saveDB=optx.pop('saveDB',False)
	if not saveDB:
		return d
	try:
		if m2pTF:
			from mongo2psql import mongo2psql
			mongo2psql(tablename,dbname)
			sys.stderr.write("===Running mongo2psql()...{dbname}::{tablename}".format(**locals()))
	except Exception as e:
		sys.stderr.write("**ERROR: {}:{}:{}\n".format("yh_quote_curr()","mongo2psql ...",str(e)))

	# Save yh_quote_curr To yh_quote_hist
	try:
		dbM = conn2mgdb(dbname=dbname)
		tablename = tablename.replace('curr','hist')
		ret=dbM[tablename].insert_many(d.to_dict(orient='records'),ordered=False)
		sys.stderr.write("===Saving history: {dbname}::{tablename}".format(**locals()))
	except Exception as e:
		sys.stderr.write("**ERROR: {}:{}:{}\n".format("yh_quote_curr()","MDB saving...",str(e)))
	return d

def qS_keyStatistics(tkLst=[],tablename='qS_keyStatistics',saveDB=False,debugTF=True,**optx):
	''' pull defaultKeyStatistics from yahoo and save to qS_keyStatistics
	e.g.,
	python3 -c "from yh_chart import qS_keyStatistics as yks;yks(['AAPL']);"
	
	'''
	if tkLst is None or len(tkLst)<1:
		tkLst = list(pull_act_tickers()['ticker'])
	dbM=None
	pbdt=datetime.datetime.now()
	jdM=yh_batchTypes(tkLst,types='quoteSummary',modules='defaultKeyStatistics',debugTF=debugTF)
	for jx in jdM:
		jx['pbdt']=pbdt
	if saveDB is True:
		zpk=optx.pop('zpk',{"ticker"})
		m,dbM,err=upsert_mdb(jdM,clientM=dbM,tablename=tablename,zpk=zpk,**optx)
		sys.stderr.write("=== {} of {} saved to {}\n".format(tkLst,jdM[-1],tablename))
	return jdM

def yh_financials_batch(tkLst=[],**optx):
	''' pull financials from yahoo and save to 'tablename'
	e.g.,
	python3 -c "from yh_chart import yh_financials_batch as yf;yf(['AAPL']);"
	OR
	python3 -c "from yh_chart import yh_financials_batch as yf;yf(modules='incomeStatementHistoryQuarterly',tablename='yh_IS_Q',zpk={'ticker','pbdate'});"
	'''
	useDB=optx.pop('useDB',False)
	tablename=optx.pop('tablename','')
	modules=optx.pop('modules','incomeStatementHistory')
	if len(tablename)>0 or useDB is True:
		dbname=optx.pop('dbname','ara')
		dbM=conn2mgdb(dbname=dbname)
		if useDB is True and (tkLst is None or len(tkLst)<1):
			tkLst=dbM['yh+'+modules].distinct("ticker")
	else:
		dbM=None
	if tkLst is None or len(tkLst)<1:
		tkLst = list(pull_act_tickers()['ticker'])
	sys.stderr.write("===Batch list:{}, tablename: {}, useDB:{}\n".format(tkLst,tablename,useDB))
	dd=[]
	for tk in tkLst:
		try:
			jd = yh_financials(tk,modules=modules,tablename=tablename,clientM=dbM,useDB=useDB,**optx)
		except Exception as e:
			sys.stderr.write("**ERROR:{}:{} @{}\n".format(tk,str(e),'yh_financials'))
			continue
		dd.extend(jd)
	return dd

def epoch2ymd(eptime):
	return int(datetime.datetime.fromtimestamp(eptime).strftime('%Y%m%d'))

# individual run of yh_financials_batch
def yh_financials(ticker,modules='incomeStatementHistory',saveDB=False,clientM=None,useDB=False,dbname='ara',tablename='',**optx):
	''' pull financials from yahoo and save to 'tablename'
	e.g.,
	python3 -c "from yh_chart import yh_financials as yf;yf('AAPL');"
	
	'''
	debugTF=getKeyVal(optx,'debugTF',False)
	types=getKeyVal(optx,'types','quoteSummary')
	xmod=modules
	jdM=[]
	if useDB is True:
		jdTmp,dbM,err=find_mdb({"ticker":ticker},clientM,dbname=dbname,tablename='yh_'+xmod)
		if len(jdTmp)<1:
			return []
		jdMod = jdTmp[0]
	else:
		try:
			jdTmp=yh_quoteSummary(ticker,modules=xmod,**optx)
			jdMod = jdTmp['quoteSummary']['result'][0][xmod]
			jD=proc_summaryQuote(jdMod,ticker=ticker,xmod=xmod)
			if jD:
				jdM=[jD] if not isinstance(jD,list) else jD
		except:
			sys.stderr.write("**ERROR:{} not found via {}\n".format(ticker,modules))
	if saveDB is True and len(tablename)>0 and len(jdM)>0:
		m,dbM,err=insert_mdb(jdM,clientM=clientM,tablename=tablename,**optx)
		sys.stderr.write("=== {} of {} saved to {}\n".format(ticker,jdM[-1],tablename))
		sys.stderr.write("{}\n".format(pd.DataFrame(jdM)[['ticker','module']].to_string(index=False)))
	return jdM

def yh_quote_comparison(tkLst=None,types='quote',sortLst=['changePercent'],rawTF=True,screenerTF=False,dfTF=False,debugTF=False,ascendingTF=False,**optx):
	'''
	show a list of ticker change / volume
	default for sector/industry performance ranking like screener
	'''
	if tkLst is None or len(tkLst)<1:
		tkLst = ['XLB','XLE','XLF','XLI','XLK','XLP','XLRE','XLU','XLV','XLY','XTL']
	elif isinstance(tkLst,str):
		tkLst = [tkLst]
	if debugTF:
		sys.stderr.write("===yh_quote_comparison() LOCALS:\n{}".format(locals()))
	df=yh_hist_query(tkLst,types=types,rawTF=rawTF,screenerTF=screenerTF,debugTF=debugTF,dfTF=dfTF,**optx)
	if not dfTF:
		return df
	df = pd.DataFrame(data=df)
	newNames={"symbol":"ticker","regularMarketPrice":"close","regularMarketChange":"change","regularMarketChangePercent":"changePercent","regularMarketOpen":"open","regularMarketDayHigh":"high","regularMarketDayLow":"low","regularMarketVolume":"volume"}
	df.rename(newNames,axis='columns',inplace=True)
	if sortLst and len(sortLst)>0 and sortLst[0] in df:
		df = df.sort_values(by=sortLst,ascending=ascendingTF)
		df.reset_index(drop=True,inplace=True)
	return df

def yh_quoteSummary(ticker,modules='summaryProfile',debugTF=False,**optx):
	if modules=='*':
		modules="""assetProfile,balanceSheetHistory,balanceSheetHistoryQuarterly,calendarEvents,cashflowStatementHistory,cashflowStatementHistoryQuarterly,defaultKeyStatistics,earnings,earningsHistory,earningsTrend,financialData,fundOwnership,incomeStatementHistory,incomeStatementHistoryQuarterly,indexTrend,industryTrend,insiderHolders,insiderTransactions,institutionOwnership,majorDirectHolders,majorHoldersBreakdown,netSharePurchaseActivity,price,quoteType,recommendationTrend,secFilings,sectorTrend,summaryDetail,summaryProfile,symbol,upgradeDowngradeHistory,fundProfile,topHoldings,fundPerformance"""
	urx = 'https://query2.finance.yahoo.com/v11/finance/quoteSummary/{ticker}?modules={modules}'
	url=urx.format(ticker=ticker,modules=modules)
	if debugTF is True:
		sys.stderr.write("==URL:\n{}\n".format(url))
	ret = requests.Session().get(url,headers=headers)
	#ret = requests.get(url,timeout=3)
	jdTmp = ret.json()
	return jdTmp

def yh_batchRaw(tkLst=[],filter='*',types='spark',ranged='5m',interval='1m',debugTF=False,**optx):
	'''
	Pulling finance.yahoo data based on the 'types' of [spark|quote|chart]
	'''
	if tkLst is None or len(tkLst)<1:
		return {}
	if types in ['spark','quote']:
		utmp = 'https://query1.finance.yahoo.com/v7/finance/{}?corsDomain=finance.yahoo.com&.tsrc=finance&'
		urx = utmp + 'symbols={}&range={}&interval={}'
		url = urx.format(types,','.join(tkLst),ranged,interval)
	else:
		utmp = 'https://query1.finance.yahoo.com/v8/finance/{}/{}?corsDomain=finance.yahoo.com&.tsrc=finance&'
		urx = utmp + 'range={}&interval={}'
		tkX = tkLst[0] if isinstance(tkLst,(list,tuple)) else tkLst
		url = urx.format(types,tkX,ranged,interval)
	includeTimestamps = optx['includeTimestamps'] if 'includeTimestamps' in optx else True
	includePrePost = optx['includePrePost'] if 'includePrePost' in optx else False
	url += "&includeTimestamps={}&includePrePost={}".format(includeTimestamps,includePrePost)
	if filter != '*' and types=='quote':
		ftx="&fields={}".format(filter)
		url += ftx
	elif filter != '*':
		ftx="&indicators={}".format(filter)
		url += ftx
	if debugTF is True:
		sys.stderr.write("==URL:\n{}\n".format(url))
	res=requests.Session().get(url,headers=headers)
	#res=requests.get(url,timeout=5)
	if debugTF:
		sys.stderr.write("=====URL STATUS:{}\n{}\n{}\n".format(res.status_code,tkLst,url))
	if res.status_code!=200:
		return {}
	try:
		jdTmp = res.json()
	except Exception as e:
			sys.stderr.write("**ERROR: {}:{}:{}\n".format("yh_batchRaw()",tkLst,str(e)))
			return {}
	return jdTmp

def batchSpark_quoteConv(jdLst=[],debugTF=False,**optx):
	colX=getKeyVal(optx,'colX',[])
	dicX={"symbol":"ticker","regularMarketPrice":"close","regularMarketChange":"change","regularMarketChangePercent":"changePercent","regularMarketOpen":"open","regularMarketDayHigh":"high","regularMarketDayLow":"low","regularMarketVolume":"volume"}
	def quote_dx2dd(jdTmp,dicX,colX):
		if "regularMarketTime" not in jdTmp:
			sys.stderr.write("**WARNING:{} {}\n".format(ticker,"is invalid."))
			return {}
		renameDict(jdTmp,dicX)
		#keep "regularMarketPreviousClose" for backward compatibility
		if "regularMarketPreviousClose" in jdTmp:  
			jdTmp['xclose']=jdTmp["regularMarketPreviousClose"]
		if "changePercent" in jdTmp:
			jdTmp['pchg']=jdTmp["changePercent"]/100.0
		epoch = int(jdTmp["regularMarketTime"])
		jdTmp['epochs']=epoch*1000
		jdTmp['pbdt']=pbdt=datetime.datetime.fromtimestamp(epoch)
		jdTmp['hhmm']=pbdt.strftime('%H%M')
		jdTmp['pbdate']=int(pbdt.strftime('%Y%m%d'))
		dd = subDict(jdTmp,colX) if len(colX)>0 else jdTmp
		return dd

	jdM=[]
	for jdTmp in jdLst:
		ticker = jdTmp['symbol']
		try:
			dd = quote_dx2dd(jdTmp,dicX,colX)
			if len(dd)<1:
				continue
			jdM.append(dd)
		except Exception as e:
			sys.stderr.write("**ERROR:{}, {}\n".format(ticker,str(e)))
			continue
	return jdM

def batchSpark_sparkConv(jdLst=[],debugTF=False,**optx):
	def spark_dx2dd(jdX):
		ticker = jdX['symbol']								
		jdd = jdX['response'][0]
		jde={}
		xClose=xChartClose=0
		if 'timestamp' not in jdd:
			return jde,xClose,xChartClose,''
		if 'meta' in jdd and 'chartPreviousClose' in jdd['meta']:
			xChartClose = jdd['meta']['chartPreviousClose']
		if 'meta' in jdd and 'previousClose' in jdd['meta']:
			xClose = jdd['meta']['previousClose']
		try:
			gap=jdd['meta']['dataGranularity']
			epoch = jdd['timestamp']
			close=jdd['indicators']['quote'][0]['close']
			pbdt = [ datetime.datetime.fromtimestamp(int(x)) for x in epoch ]
			epochs = [ x*1000 for x in epoch ]
			hhmm = [x.strftime('%H%M') for x in pbdt]
			pbdate = [int(x.strftime('%Y%m%d')) for x in pbdt]
			ne = len(epoch)
		except Exception as e:
			sys.stderr.write("**ERROR:{}, {}\n".format(ticker,str(e)))
			return jde, xClose,xChartClose,gap
		jde['ticker']=[ticker]*ne
		jde['close']=close
		jde['epochs']=epochs
		jde['pbdt']=pbdt
		jde['hhmm']=hhmm
		jde['pbdate']=pbdate
		return jde,xClose,xChartClose,gap
	jdM=[]
	for jdX in jdLst:
		jde,xClose,xChartClose,gap = spark_dx2dd(jdX)
		if len(jde)<1:
			continue
		else:
			dx = pd.DataFrame(data=jde)
			dx = dx.dropna()
			# NOTE, 
			# Minute data 'change' is calc based on last-day close instead of 'gap' interval
			# This can be be a potential bug if range is more than 1-d
			if gap[-1] in ['m'] and xClose>0: # for minute data with '1d' range ONLY
				dx['xclose'] = xClose
				dx['change'] = dx['close'] - dx['xclose']
				dx['pchg'] = dx['change']/dx['xclose']
			elif gap[-1] not in ['m'] and xChartClose>0: # for minute data
				dx['xclose'] = dx['close'].shift()
				dx.loc[0,'xclose'] = xChartClose
				dx['change'] = dx['close'] - dx['xclose']
				dx['pchg'] = dx['change']/dx['xclose']
			else:
				dx['xclose'] = dx['pchg'] = dx['change'] = 0.0
			dd = dx.to_dict(orient='records')
			jdM.extend(dd)
	return jdM

def chart2df(jdLst=[],**optx):
	#- get DF
	jdTmp=jdLst[0]
	dx=pd.DataFrame(jdTmp['indicators']['quote'][0])

	# get timestamp
	pbts = np.array(jdTmp['timestamp'])
	pbdt = np.vectorize(datetime.datetime.fromtimestamp)(pbts)
	dx['pbdt'] = pbdt
	dx['pbdate'] = [int(x.strftime("%Y%m%d")) for x in pbdt]
	dx['epochs'] = pbts * 1000

	# get ticker
	dx.loc[:,'ticker']=jdTmp['meta']['symbol']

	# Use previousClose to calc change and pchg (changePercent)
	# NOTE, 
	# Minute data 'change' is calc based on last-day close instead of 'gap' interval
	# This can be be a potential bug if range is more than 1-d
	gap=jdTmp['meta']['dataGranularity']
	if gap[-1]=='m': # for minute data
		dx.loc[:,'xclose']=jdTmp['meta']['previousClose']
	else:
		xclose=jdTmp['meta']['chartPreviousClose']
		dx.loc[:,'xclose']=dx['close'].shift()
		dx.loc[0,'xclose'] = xclose

	# build date/time as pbdate/epochs column into datafame
	if 'indicators' in jdTmp and 'adjclose' in jdTmp['indicators']:
		adjclose = np.array(jdTmp['indicators']['adjclose'][0]['adjclose'])
		dx['adjusted'] = adjclose
	# remove NA rows related to [close] data
	dx.dropna(subset=['close'],inplace=True)
	dx['change']=dx['close']-dx['xclose']
	dx['pchg']=dx['change']/dx['xclose']
	return dx.to_dict(orient='records')

def yh_batchSpark(tkLst=[],filter='*',types='spark',nchunk=20,saveDB=True,dbname='ara',tablename='yh_spark_hist',zpk={'ticker','epochs'},t2='yh_quote_curr',t2zpk={'ticker'},isEoD=False,**optx):
	'''
	pull data from types=[spark|quote] then save to mgDB 'dbname':: 'tablename' and 't2' respectively
	Note,
	if isEoD=True: quote data save to  both mgDB 'dbname':: 'tablename' and 't2'
	if tablename or t2 ='' , nothing will be saved in 'tablename' or 't2'
	quote data will also be saved to pgDB 'dbname'::'t2' if isEoD=True and types='quote'
	'''
	debugTF=getKeyVal(optx,'debugTF',False)
	if debugTF:
		sys.stderr.write("===LOCALS: {}\noptx: {}\n".format(locals(),optx))
	dbM=conn2mgdb(dbname=dbname)
	if tkLst is None or len(tkLst)<1:
		tkLst = list(pull_act_tickers()['ticker'])
	elif isinstance(tkLst,str):
		tkLst = [tkLst]
	chunkLst = list2chunk(tkLst,nchunk)
	jdN=[]
	colX=['ticker','open','high','low','close','xclose','volume','pchg','change','pbdate','epochs','pbdt','hhmm']
	for j,tkM in enumerate(chunkLst):
		jdTmp = yh_batchRaw(tkM,types=types,**optx)
		jdQC = {}

		if types in ['spark','chart']:
			try:
				jdLst = jdTmp[types]['result']
			except Exception as e:
				continue
			if types=='chart':
				jdM = chart2df(jdLst,**optx)
			else:
				jdM = batchSpark_sparkConv(jdLst,**optx)
			if len(jdM)<1:
				continue
			if saveDB is True:
				m,dbM,err=insert_mdb(jdM,clientM=dbM,tablename=tablename,zpk=zpk,**optx)
				if debugTF:
					sys.stderr.write("++optx: {}\n".format(optx))
					sys.stderr.write("=== {} of {} saved to {}\n".format(tkM,jdM[-1],tablename))
			jdN.extend(jdM)
		else: # for case of types.lower()=='quote'
			jdLst = jdTmp['quoteResponse']['result']
			jdQC = batchSpark_quoteConv(jdLst,**optx)
			if len(jdQC)<1:
				continue
			jdM = subDict(jdQC,colX)
			if saveDB is True and len(jdQC)>0:
				if len(t2)>0:
					for tname in t2.split(','):
						qc,dbM,erq=upsert_mdb(jdQC,clientM=dbM,tablename=tname,zpk=t2zpk,**optx)
						sys.stderr.write("=== {} of {} saved to {}\n".format(tkM,jdQC[-1],tname))
				if isEoD is True:
					m,dbM,err=insert_mdb(jdM,clientM=dbM,tablename=tablename,zpk=zpk,**optx)
					sys.stderr.write("=== {} of {} saved to {}\n".format(tkM,jdM[-1],tablename))
			jdN.extend(jdQC)
		if debugTF:
			sys.stderr.write("=== tkM:{}[{}/{}], last:\n{}\n".format(tkM,len(jdM),len(jdN),jdN[-1]))

	if saveDB is True and len(t2)>0 and len(jdN)>0 and types.lower()=='quote' and isEoD is True:
		try:
			df = pd.DataFrame(jdN)
			df = df.drop(['_id'], axis=1)
			if debugTF:
				sys.stderr.write("=== Save to pgDB::{} of {} rows(1st-last)\n{}\n{}\n".format(t2,len(df),df.iloc[:1],df.iloc[-1:]))
			from _alan_calc import save2pgdb
			for tname in t2.split(','):
				save2pgdb(df,dbname,tablename=tname)
		except Exception as e:
			sys.stderr.write("**ERROR: {}:{}\n".format("save2pgdb",str(e)))
	return jdN

def proc_raw_fmt(dc):
	sys.stderr.write("===proc_raw_fmt {}:{}".format(type(dc),dc)[:100]+"\n")
	if isinstance(dc,list):
		for jk,jdc in enumerate(dc):
			dc[jk]=proc_raw_fmt(jdc)
		return dc
	elif not isinstance(dc,dict):
		return {}
	jdc = {}
	for ky,val in dc.items():
		if ky == 'maxAge':
			continue
		elif not val:
			continue
		cv=val
		if isinstance(val, list):
			cv=proc_raw_fmt(val)
		elif isinstance(val, dict):
			if 'raw' in val:
				cv=val['raw']
		jdc.update({ky:cv}) 
		
	return jdc

def proc_summaryQuote(modObj={},xmod='',ticker=''):
	if not all([modObj,xmod,ticker]):
		return {}
	outLst= proc_raw_fmt(modObj)
	ktmp=list(outLst)
	if len(ktmp)==1 and isinstance(outLst[ktmp[0]],list):
		ymod=ktmp[0]
		outLst=outLst[ymod]
		for j,jx in enumerate(outLst):
			if 'endDate' in outLst[j] and outLst[j]['endDate']>10**9:
				outLst[j]['pbdate']= epoch2ymd(outLst[j]['endDate'])
			elif 'quarter' in outLst[j] and outLst[j]['quarter']>10**9:
				outLst[j]['pbdate']= epoch2ymd(outLst[j]['quarter'])
			elif 'reportDate' in outLst[j] and outLst[j]['reportDate']>10**9:
				outLst[j]['pbdate']= epoch2ymd(outLst[j]['ReportDate'])
			outLst[j]['module']=xmod
			outLst[j]['ticker']=ticker
			if ymod:
				outLst[j]['submodule']=ymod
	else:
		ymod=None
		kyLst=list(outLst.keys())
		for ky in kyLst:
			outLst['module']=xmod
			outLst['ticker']=ticker
	return outLst

def process_summaryRawData(dc={}):
	dd={}
	if not isinstance(dc,dict):
		return {}
	for ky,va in dc.items():
		if isinstance(va,dict):
			if 'raw' in va:
				dd.update({ky:va['raw']})
		else:
			if va is not None:
				dd.update({ky:va})
	if 'endDate' in dd and dd['endDate']>10**9:
		dd['pbdate']= epoch2ymd(dd['endDate'])
	elif 'quarter' in dd and dd['quarter']>10**9:
		dd['pbdate']= epoch2ymd(dd['quarter'])
	return dd

def yh_batchTypes(tkLst=[],filter='*',types='quote',nchunk=20,modules='summaryProfile',debugTF=False,**optx):
	#xlst = (unicode, str) if hasattr(__builtins__,"unicode") else str
	#if isinstance(tkLst,xlst):
	#	chunkLst = [tkLst]
	if types in ['quote','spark']:
		chunkLst = list2chunk(tkLst,nchunk)
	else:
		chunkLst = tkLst
	jdM=[]
	for j,tkM in enumerate(chunkLst):
		try:
			if types=='quoteSummary': # tkM is a string of ticker name
				jdTmp = yh_quoteSummary(tkM,modules=modules,debugTF=debugTF,**optx)
			else: # tkM is a list of tickers
				jdTmp = yh_batchRaw(tkM,filter=filter,types=types,debugTF=debugTF,**optx)
			if not isinstance(jdTmp,dict) or len(jdTmp)<1:
				continue
			if types=='quoteSummary': # tkM is a ticker string
				try:
					jdLst = jdTmp['quoteSummary']['result'][0]
				except:
					continue
				for xmod in jdLst:
					modObj=jdLst[xmod]
					jD=proc_summaryQuote(modObj,ticker=tkM,xmod=xmod)
				jdX=[jD] if not isinstance(jD,list) else jD
			elif types=='quote':
				jdX = jdTmp['quoteResponse']['result']
			else:
				jdX = jdTmp[types]['result']
			if debugTF:
				sys.stderr.write("==={}.{}\njdTmp\n{}\njdX\n{}\n".format(j,tkM,jdTmp,jdX[-1]))
		except Exception as e:
			sys.stderr.write("**ERROR: {}:{}\n".format(types,str(e)))
			continue
		jdM.extend(jdX)
	if debugTF:
		sys.stderr.write("===jdM:\n{}".format(jdM))
	return jdM

# DEPRECATED
def raw2screener_output_1(jobj):
	d = dict(ticker=jobj['symbol'])
	d.update(changePercent=jobj['regularMarketChangePercent'])
	if 'marketCap' in jobj:
		d.update(marketCap=jobj['marketCap'])
	else:
		d.update(marketCap=0)
	for xk,ky in zip(['ytdReturn','regularMarketVolume','regularMarketPrice','regularMarketChange','shortName'],['ytdReturn','volume','price','change','company']) :
		if xk in jobj:
			d.update({ky:jobj[xk]})
	return d

def raw2spark_output(jobj):
	from _alan_calc import renameDict
	colx=['ticker','close','changePercent','regularMarketPreviousClose','epochs','pbdt','change']
	ds={x:y for x,y in jobj.items() if x in colx}
	renameDict(ds,{'regularMarketPreviousClose':'xclose','changePercent':'pchg'})
	ds['hhmm']=ds['pbdt'].strftime('%H%M')
	ds['pchg'] /= 100.0
	return ds

def yh_hist_query(tkLst=[],filter='*',types='quote',nchunk=20,rawTF=False,screenerTF=False,dfTF=False,debugTF=False,dbname=None,tablename=None,**optx):
	'''
	Pull minute ohlc pricing data from Yahoo but use marketVolume as volume
	since market data has 15-minute delay, latest 15 marketVolumes become 0 
	'''
	if len(tkLst)<1:
		tkLst = list(pull_act_tickers()['ticker'])
	jdLst = yh_batchTypes(tkLst,filter=filter,types=types,nchunk=nchunk,debugTF=debugTF,**optx)
	colX = ["ticker","open","high","low","close","volume","xclose","change","pchg","epochs",'hhmm',"pbdate","pbdt"]
	dLst=[]
	df=pd.DataFrame()
	clientM=None
	tablename = 'yh_{}_temp'.format(types) if tablename is None else tablename
	for j,jdTmp in enumerate(jdLst):
		try:
			jdX={}
			if 'response' in jdTmp:
				ticker = jdTmp['symbol']
				jdd = jdTmp['response'][0]
			elif 'meta' in jdTmp:
				ticker = jdTmp['meta']['symbol']
				jdd = jdTmp
			else: #- for 'quote' parsing
				if "regularMarketPrice" not in jdTmp:
					continue
				if "regularMarketTime" not in jdTmp:
					continue
				jdTmp['epochs']=jdTmp['regularMarketTime']*1000
				jdTmp['pbdt'] = datetime.datetime.fromtimestamp(jdTmp['regularMarketTime'])
				jdTmp['pbdate'] = int(jdTmp['pbdt'].strftime('%Y%m%d'))
				newNames={"symbol":"ticker","regularMarketPrice":"close","regularMarketChange":"change","regularMarketChangePercent":"changePercent","regularMarketOpen":"open","regularMarketDayHigh":"high","regularMarketDayLow":"low","regularMarketVolume":"volume","regularMarketPreviousClose":"xclose","regularMarketTime":"epoch"}
				if rawTF:
					renameDict(jdTmp,mapper=newNames)
					if debugTF:
						sys.stderr.write("{}\n".format(jdTmp))
					if screenerTF==True:
						colx=["change","changePercent","company","marketCap","close","ticker","volume","epochs","pbdt"]
						#ds=raw2screener_output_1(jdTmp)
						ds =  subDict(jdTmp,colx)
						renameDict(ds,{"close":"price"})
					elif screenerTF>0: # original False case
						colx=list( set(newNames.values()).union(['epochs','pbdt','hhmm','pbdate','marketCap']) )
						ds =  subDict(jdTmp,colx)
					else:
						ds = jdTmp 
					if all([dbname,tablename]):
						zpk=getKeyVal(optx,'zpk',['ticker','epochs'])
						#mobj,clientM,err_msg = write2mdb(ds,clientM,dbname=dbname,tablename=tablename,zpk=zpk)
						mobj,clientM,err_msg = insert_mdb(ds,clientM,dbname=dbname,tablename=tablename,zpk=zpk)
						if debugTF:
							sys.stderr.write("{}\nSave to {}::{}\n".format(ds,dbname,tablename))
					dLst.append(ds)
					continue
				#- for 'spark' and 'chart' parsing
				dx=pd.DataFrame([jdTmp])
				dx.rename(newNames,axis='columns',inplace=True)
				if 'volume' not in dx:
					continue
				dx.dropna(subset=['volume'],inplace=True)
				if len(dx)<1:
					continue
				colX = [x for x in colX if x in dx.columns]
				dm = dx[colX]
				if debugTF:
					sys.stderr.write("quote:\n{}".format(dm.tail()))
				df= pd.concat([df,dm])
				continue
			xClose=None
			if 'meta' in jdd and 'previousClose' in jdd['meta']:
				xClose = jdd['meta']['previousClose']
			epoch = jdd['timestamp']
			for x,y in jdd['indicators']['quote'][0].items():
				jdX[x] = y
			jdX['epochs'] = np.array(epoch)*1000
			dx=pd.DataFrame(jdX)
			dx['ticker']=ticker
			if 'pchgTF' in optx and optx['pchgTF'] is False:
				df= pd.concat([df,dx])
				continue
			elif 'pchgTF' in optx and optx['pchgTF'] and jdd['meta']['dataGranularity'][:1]!='m':
				dx['pchg'] = dx['close'].pct_change()
				dx['change'] = dx['close'].diff()
				xChartClose = jdd['meta']['chartPreviousClose']
				dx.loc[dx.index[0],'pchg'] = dx.loc[dx.index[0],'close']/xChartClose - 1
				dx.loc[dx.index[0],'change'] = dx.loc[dx.index[0],'close']-xChartClose

			pbdt = [ datetime.datetime.fromtimestamp(int(x)) for x in epoch ]
			dx['hhmm']=[x.strftime('%H%M') for x in pbdt]
			dx['pbdate']=[x.strftime('%Y%m%d') for x in pbdt]
			dx['pbdate']=dx['pbdate'].astype(int)
			dx['pbdt']=pbdt
			dx = dx.dropna()
			if xClose is not None and xClose>0:
				dx['pchg'] = dx['close']/xClose -1
				dx['change'] = dx['close'] - xClose
				dx['xclose'] = xClose
			colX = [x for x in colX if x in dx.columns]
			dm = dx[colX]
			if debugTF:
				sys.stderr.write("{}".format(dm.tail()))
			if dfTF:
				df= pd.concat([df,dm])
			else:
				dLst.extend(dm.to_dict(orient='records'))
		except Exception as e:
			sys.stderr.write("**ERROR: {}:{}:{}\n".format(j,jdTmp,str(e)))
			continue
	if len(df)>0:
		df.reset_index(drop=True,inplace=True)
	if len(dLst)>0:
		if dfTF:
			dLst = pd.DataFrame(dLst)
		return dLst
	return df

def func2mdb(tkLst,tablename='yh_spark_hist',dbname='ara',funcN='yh_hist_query',zpk={'ticker','hhmm'},debugTF=False,**optx):
	'''
	Run 'funcN'() and save the result to mongoDB
	'''
	from _alan_str import write2mdb,insert_mdb
	if funcN in globals():
		funcArg=globals()[funcN]
	else:
		return {}
	df = funcArg(tkLst,debugTF=debugTF,dbname=dbname,**optx)
	try:
		#dbM = conn2mgdb(dbname=dbname)
		#ordered=False # to insert whatever is available
		#if not tablename in dbM.collection_names():
		#	zsrt=[1]*len(zpk)
		#	dbM[tablename].create_index([(k,v) for k,v in zip(zpk,zsrt)],unique=True)
		#ret=dbM[tablename].insert_many(df,ordered=ordered)
		dbM=None;zpk=zpk
		mobj,dbM,err_msg =upsert_mdb(df,dbM,dbname=dbname,tablename=tablename,zpk=zpk,insertOnly=True)
	except Exception as e:
		sys.stderr.write("**ERROR: {}:{}\n".format("func2mdb()",str(e)))
	return df	
def renewChk(pbdtCurr,pbdtMod,deltaTolerance=86400):
	deltaPassed=pd.Timedelta(pbdtCurr - pbdtMod).total_seconds()
	sys.stderr.write(" --curr:{},last:{}:deltaPassed:{}\n".format(pbdtCurr,pbdtMod,deltaPassed))
	return deltaPassed>deltaTolerance

def lastRunChk(objChk={},tableChk='',deltaTolerance=43200,clientM=None,dbname='ara',**optx):
	pbdtCurr=pd.datetime.now()
	lastObj,clientM,_=find_mdb(objChk,clientM=clientM,dbname=dbname,tablename=tableChk,limit=1)
	if not lastObj:
		pbdtMod=pbdtCurr
		renewTF=True
	else:
		pbdtMod=lastObj[0]['pbdt']
		renewTF=renewChk(pbdtCurr,pbdtMod,deltaTolerance)
		if renewTF:
			pbdtMod=pbdtCurr
	objChk.update(pbdt=pbdtMod)
	return renewTF,objChk,clientM

def batchOTF(funcArg,tkLst=[],tableChk='',zpkChk=["ticker"],deltaTolerance=43200,**optx):
	for ticker in tkLst: 
		sys.stderr.write("==Batch Running:{} on {}\n".format(ticker,funcArg))
		runOTF(funcArg,ticker,tableChk,zpkChk=zpkChk,deltaTolerance=deltaTolerance,**optx)

@safeRunArg([])
def runOTF(funcArg,ticker='',tableChk='',zpkChk=["ticker"],deltaTolerance=43200,**optx):
	'''
	To  ticker='AAPL'
	where real-time data is only grab based on 'deltaTolerance' in seconds
	current setup is half-days
	'''

	if isinstance(ticker,list):
		tkLst=ticker
		dLst=[]
		for tkX in tkLst: 
			sys.stderr.write("==BATCH Running:{} on {}\n".format(ticker,funcArg))
			dd = runOTF(funcArg,tkX,tableChk,zpkChk=zpkChk,deltaTolerance=deltaTolerance,**optx)
			if isinstance(dd,list):
				dLst.extend(dd)
			else:
				dLst.extend([dd])
		return dLst

	if isinstance(funcArg,str):
		if funcArg in globals() and hasattr(globals()[funcArg],'__call__'):
			funcArg =  globals()[funcArg]
		else:
			return []
	sys.stderr.write("==START Running:{} on {}\n".format(ticker,funcArg))
	dbname=getKeyVal(optx,'dbname','ara')
	optx.update({'dbname':dbname})
	tablename=getKeyVal(optx,'tablename','')
	if not all([tablename, ticker]):
		return []
		
	if 'modules' in optx:
		objChk=dict(ticker=ticker,module=optx['modules'])
	else:
		objChk=dict(ticker=ticker)
	if not tableChk:
		tableChk=tablename+'_chk'
	renewTF,objChk,clientM = lastRunChk(objChk=objChk,tableChk=tableChk,deltaTolerance=deltaTolerance,**optx)
	if renewTF:
		sys.stderr.write("==Data outdated or never run, Running:{}\n".format(funcArg))
		retObj = funcArg(ticker,**optx)
		if len(retObj)<1:
			return []

		xChk={ky:val for ky,val in retObj[0].items() if ky in zpkChk}
		objChk.update(xChk)
		retObj,clientM,errChk = upsert_mdb(retObj,**optx)
		sys.stderr.write(" --Update {} to {}\n".format(objChk,tableChk))
		objChk,clientM,errChk = upsert_mdb(objChk,clientM=clientM,tablename=tableChk,zpk=zpkChk)
	else:
		sys.stderr.write("==Data exist, LoadFromTable:{}\n".format(tablename))
		objChk.pop('pbdt',None)
		optx.get('zpk',None)
		retObj,clientM,errMsg = find_mdb(objChk,clientM=clientM,**optx)
	return retObj

def main_tst():
	args = sys.argv[1:]
	if len(args)<1:
		#exit(0)
		args=['AAPL','IBM']
	elif len(args)==1 and ',' in args[0]:
		args = args[0].split(',')
	elif len(args)==1 and '-' in args[0]:
		args = sys.stdin.read().strip().split()
	#df = yh_spark_hist(args)
	#df = yh_hist_query(args,types='spark',ranged='1d',pchgTF=False)
	#df = yh_hist_query(args,types='quote',pchgTF=False,filter='symbol,regularMarketPrice,regularMarketChange,regularMarketOpen,regularMarketDayHigh,regularMarketDayLow,regularMarketVolume')
	#df = yh_hist_query(args,types='quote')
	df = yh_quote_curr(args)
	if isinstance(df,list):
		sys.stderr.write("\n{}".format(df))
	else:
		sys.stderr.write("\n{}".format(df.to_string()))

if __name__ == '__main__':
	main_tst()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
#- Description:
Update sp500_component table in pgDB
Update dow_component table in pgDB
Update nasdaq_ipos table in MDB
Update madmoney_hist table in MDB
#- Available methods:
  sp500_component dow_component madmoney_hist nasdaq_ipos
#- Usage of:
python3 theme_list.py sp500_component
python3 theme_list.py dow_component
python3 theme_list.py nasdaq_ipos
python3 theme_list.py madmoney_hist --extra_xs='airdate="07"'
#- OR
python3 theme_list.py nasdaq_ipos --extra_xs='month="2020-07"'
OR check ipo for the current month
python3 theme_list.py nasdaq_ipos 
#- OR
python3 theme_list.py madmoney_hist --extra_xs='airdate="2020-02-15"'
#- OR for the last 180 days
python3 theme_list.py madmoney_hist --extra_xs='airdate="180"'
#- OR for the last 7 days
python3 theme_list.py madmoney_hist --extra_xs='airdate="07"'
#- OR for ALL history
python3 theme_list.py madmoney_hist --extra_xs='airdate="%"'

#- Methods available :
def sp500_component
def dow_component
def madmoney_hist
def nasdaq_ipos
def madmoney_screener (DEPRECATED)
def spy_component (DEPRECATED)
def nasdaq_spos (DEPRECATED)
def spy_component (DEPRECATED)

#- Ref Sites:
https://en.wikipedia.org/wiki/List_of_S%26P_500_companies
https://madmoney.thestreet.com/screener/index.cfm?stocks&showrows=2000
https://us.spdrs.com/site-content/xls/SPY_All_Holdings.xls?fund=SPY&docname=All+Holdings&onyx_code1=&onyx_code2=
https://www.nasdaq.com/markets/ipos/activity.aspx?tab=filings&month=2019-07
https://api.nasdaq.com/api/ipo/calendar?date=2019-10

Last mod., 
Fri Jul 10 11:58:36 EDT 2020
'''
import sys
import pandas as pd
from optparse import OptionParser
from sqlalchemy import create_engine
import datetime
import requests
import json
import re
from pymongo import MongoClient
from bs4 import BeautifulSoup, element as bs4Element
from _alan_calc import save2pgdb,conn2mgdb,renameDict
from _alan_str import write2mdb,insert_mdb,upsert_mdb

# to deal with ç u'\xe7' character
if sys.version_info.major == 2:
	reload(sys)
	sys.setdefaultencoding('utf8')
	from urlparse import parse_qs
else:
	from urllib.parse import parse_qs


def epoch2ymd(x,s=1000,fmt='%Y%m%d'):
	'''
	convert datetime from Unix epoch to YYYYMMDD 
	'''
	return datetime.datetime.fromtimestamp(int(x/s)).strftime(fmt)

def getKeyVal(opts={},key='',val=None):
	'''
	Get value from dict 'opts' default to 'val'
	Example:
	  saveDB=getKeyVal(opts,key='saveDB',val=True)
	'''
	if not all([key,opts]):
		return None
	return opts[key] if key in opts else val

def dow_component(hostname='localhost',dbname='ara',tablename='dow_component',debugTF=False,**kwargs): 
	'''
	Update pgDB Table: dow_component
	Get DOW30 list from wiki
	and save to both mongoDB and postgreSQL
	DB Table: dow_component
	ref site: https://en.wikipedia.org/wiki/Dow_Jones_Industrial_Average
	'''
	url='https://en.wikipedia.org/wiki/Dow_Jones_Industrial_Average'
	saveDB=getKeyVal(kwargs,key='saveDB',val=True)
	if debugTF:
		sys.stderr.write('URL:{url}\n'.format(url=url))
	try:
		df = pd.read_html(url,attrs={'id':'constituents'},index_col=False,header=0)[0]
		if len(df)<30:
			return {}
		df['ticker'] = [s.replace('.','-') for s in df['Symbol']]
		df.columns = [re.sub('[(), ]', '',x) for x in df.columns]
	except Exception as e:
		sys.stderr.write("**ERROR: dow_component:{}\n".format(str(e)))
		return {}
	if saveDB:
		save2pgdb(df,db=dbname,tablename=tablename)
		mobj,_,_ = insert_mdb(df,clientM=None,dbname=dbname,tablename=tablename,wmode='replace')
	return df

def sp500_component(hostname='localhost',dbname='ara',tablename='sp500_component',debugTF=False,**kwargs): 
	'''
	Update pgDB Table: sp500_component
	Get Sp500 list from wiki
	and save to both mongoDB and postgreSQL
	ref site: https://en.wikipedia.org/wiki/List_of_S%26P_500_companies
	'''
	saveDB=getKeyVal(kwargs,key='saveDB',val=True)
	url='https://en.wikipedia.org/wiki/List_of_S%26P_500_companies'
	if debugTF:
		sys.stderr.write('URL:{url}\n'.format(url=url))
	try:
		df = pd.read_html(url,attrs={'id':'constituents'},index_col=False,header=0)[0]
		if len(df)<500:
			return {}
		df['ticker'] = [s.replace('.','-') for s in df['Symbol']]
		save2pgdb(df,db=dbname,tablename=tablename)
	except Exception as e:
		sys.stderr.write('**ERROR: sp500_component:{}\n'.format(str(e)))
		return {}
	if saveDB:
		save2pgdb(df,db=dbname,tablename=tablename)
		mobj,_,_ = insert_mdb(df,clientM=None,dbname=dbname,tablename=tablename,wmode='replace')
	return df

# DEPRECATED
def spy_component(hostname='localhost',dbname='ara',tablename='spy_component',debugTF=False,**kwargs): 
	'''
	Update pgDB Table: spy_list
	get SPY list in daily basis
	ref table: spy_list
	ref site: https://us.spdrs.com/site-content/xls/SPY_All_Holdings.xls?fund=SPY&docname=All+Holdings&onyx_code1=&onyx_code2=
	'''
	url='https://us.spdrs.com/site-content/xls/SPY_All_Holdings.xls?fund=SPY&docname=All+Holdings&onyx_code1=&onyx_code2='
	if debugTF:
		sys.stderr.write('URL:{url}\n'.format(url=url))
	try:
		df = pd.read_excel(url,index_col=False,header=3)
		if len(df)<500:
			return {}
		df.dropna(subset=['Sector'],inplace=True)
		df = df.drop(df[df['Identifier']=='CASH_USD'].index)
		df = df.reset_index(drop=True)
		df['ticker'] = [s.replace('.','-') for s in df['Identifier']]
		df.at[df['ticker']=='CCL-U','ticker'] = 'CCL'
	except Exception as e:
		sys.stderr.write('**ERROR: spy_component:{}\n'.format(str(e)))
		return {}
	mobj,_,_ = write2mdb(df,clientM=None,dbname=dbname,tablename=tablename,zpk={'*'})
	return df

def add_sector_industry(df=[]):
	from yh_chart import runOTF
	from _alan_calc import subDict
	if len(df)<1:
		return []
	tkLst=df['ticker'].values
	datax=runOTF('yh_financials',list(df['ticker'].values),modules="summaryProfile",dbname='ara',tablename="yh_summaryProfile",zpk={'ticker'},deltaTolerance=8640000)
	if len(datax)<1:
		return df
	dg =  subDict(pd.DataFrame(datax),['ticker','sector','industry'])
	df = df.merge(dg,on='ticker',how='left')
	return df

def nasdaq_ipos(hostname='localhost',dbname='ara',tablename='nasdaq_ipos',tab='filing',month='',debugTF=False,**kwargs): 
	'''
	Update MDB Table: nasdaq_ipos
	get recent IPOs
	ref table: nasdaq_ipos
	ref site: https://api.nasdaq.com/api/ipo/calendar?date=2019-10
	'''
	if len(month)<1:
		month = datetime.datetime.now().strftime('%Y-%m')
	urx='https://api.nasdaq.com/api/ipo/calendar?date={month}'
	url = urx.format(month=month)
	sys.stderr.write(' --URL:{}\n'.format(url))
	headers={'Content-Type': 'application/json', 'Accept': 'application/json', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}
	try:
		#ret =requests.get(url,timeout=5)
		ret=requests.Session().get(url,headers=headers)
		jsx=ret.json()
		actions=getKeyVal(kwargs,key='actions',val='priced')
		data = jsx['data'][actions]['rows']
		if not data:
			actions='filed'
			data = jsx['data'][actions]['rows']
			if not data:
				return {}
		colX=jsx['data'][actions]['headers']
		df = pd.DataFrame(data)
		df = renameDict(df,colX)
		if debugTF:
			sys.stderr.write(' --DATA:\n{}\n'.format(data))
			sys.stderr.write(' --df:\n{}\n'.format(df))
		if "Exchange/ Market" in df:
			df=df.rename(columns={"Exchange/ Market":"Market"})
		if debugTF:
			sys.stderr.write('COLUMNS:{}\n'.format(df.columns))
		df.dropna(subset=['Symbol'],inplace=True)
		if len(df)<1:
			return {}
		df['ticker'] = [s.replace('.','-') for s in df['Symbol']]
		if 'Offer Amount' in df:
			df['offer'] = [float(s.replace('$','').replace(',','')) for s in df['Offer Amount']]
			df = df.drop(['Offer Amount'], axis=1)
		if 'Date' in df:
			df['pbdate'] = [int(datetime.datetime.strptime(s,'%m/%d/%Y').strftime('%Y%m%d')) for s in df['Date']]
		elif 'Date Filed' in df:
			df['pbdate'] = [int(datetime.datetime.strptime(s,'%m/%d/%Y').strftime('%Y%m%d')) for s in df['Date Filed']]
		elif 'Date Priced' in df:
			df['pbdate'] = [int(datetime.datetime.strptime(s,'%m/%d/%Y').strftime('%Y%m%d')) for s in df['Date Priced']]

	except Exception as e:
		sys.stderr.write('**ERROR: nasdaq_ipos:{}\n'.format(str(e)))
		return {}
	df.columns = [x.replace(' ','').lower() for x in df.columns]
	df = add_sector_industry(df)
	mobj,_,_ = upsert_mdb(df,clientM=None,dbname=dbname,tablename=tablename,zpk={'ticker','pbdate','actions'})
	return df

def nasdaq_spos(hostname='localhost',dbname='ara',tablename='nasdaq_spos',debugTF=False,**kwargs): 
	return nasdaq_ipos(hostname=hostname,dbname=dbname,tablename=tablename,debugTF=debugTF,**kwargs)

def madmoney_hist(hostname='localhost',dbname='ara',tablename='madmoney_hist',airdate='180',pricehigh=1000,pricelow=0,sortby='airdate',showrows=2000,rawTF=False,saveDB=True,debugTF=False,**kwargs):
	return madmoney_screener(hostname=hostname,dbname=dbname,tablename=tablename,airdate=airdate,pricehigh=pricehigh,pricelow=pricelow,sortby=sortby,showrows=showrows,rawTF=rawTF,saveDB=saveDB,debugTF=debugTF,**kwargs)

def madmoney_screener(hostname='localhost',dbname='ara',tablename='madmoney_hist',airdate='180',pricehigh=1000,pricelow=0,sortby='airdate',showrows=2000,rawTF=False,saveDB=True,debugTF=False,**kwargs):
	'''
	Update MDB Table: madmoney_hist
	get madmoney list in daily basis
	ref table: madmoney_hist
	ref site: https://madmoney.thestreet.com/screener/index.cfm?showview=stocks&showrows=500
	'''
	sys.stderr.write('{}\n'.format(locals()))
	urx='https://madmoney.thestreet.com/screener/index.cfm?showview=stocks&showrows={showrows}&airdate={airdate}'
	CallD={'5':'buy','4':'positive','3':'hold','2':'negative','1':'sell'}
	SegmentD=dict(F='Featured Stock',D='Discussed Stock ',C='Callers Stock',I='Guest Interview',L='Lighting Round',M='Mail Bag',G='Game Plan',S='Sudden Death')
	url= urx.format(showrows=showrows,airdate=airdate)
	jd = dict(symbol='',airdate=airdate,called='%',industry='%',sector='%',segment='%',pricelow=pricelow,pricehigh=pricehigh,sortby=sortby)
		
	sys.stderr.write('URL:\n{}\nPOST:{}\n'.format(url,jd))
	try:
		#res = requests.get(url,timeout=10)
		res = requests.post(url,data=json.dumps(jd),timeout=10)
		if res.status_code!=200:
			sys.stderr.write('**HTTP: {} @ {}\n'.format(res.status_code,url))
			return {}
		xstr=res.content
		df = pd.read_html(xstr,attrs={'id':'stockTable'},index_col=False,header=0)[0]
		if len(df)<1 or pd.isnull(df.loc[0,'Company']):
			sys.stderr.write('**{} @ {}\n'.format("No sotcks were found...",url))
			return {}
		s = BeautifulSoup(xstr,'lxml')
		trLst=s.find_all('table')[0].find_all('tr')
		dd=[]
		for j,rwx in enumerate(trLst):
			if j<1:
				tagX='th' 
				tdLst = rwx.find_all(tagX)
				vHdr = [x.text for x in tdLst]
				continue
			else:
				tagX='td' 
			tdLst = rwx.find_all(tagX)
			vLst=[]
			if len(tdLst)<len(vHdr):
				continue
			for k,xTd in enumerate(tdLst):
				if isinstance(xTd.next,bs4Element.Tag) and xTd.next.has_attr('alt'):
					xv = xTd.next['alt']
				elif isinstance(xTd.next,bs4Element.Tag) and xTd.next.has_attr('href'):
					xv =  parse_qs(xTd.next['href'])['symbol'][0]
				else:
					xv = xTd.text
				vLst.append(xv)

			dx = dict(zip(vHdr,vLst))
			dd.append(dx)
		df=pd.DataFrame(dd)
	except Exception as e:
		sys.stderr.write('**ERROR: {} @{}\n'.format(str(e),'madmoney_screener()'))
		return {}
	if rawTF:
		return df
	try:
		df['CallDscript']=[CallD[x] for x in df['Call']]
		df['SegmentDscript']=[SegmentD[x] for x in df['Segment']]
		df['Price'] = df['Price'].apply(lambda x:float(x.strip('$')))
		mmdd = [int("".join(x.split('/'))) for x in df['Date']]
		tdate=datetime.datetime.today()
		cmmdd = int(tdate.strftime("%m%d"))
		cYr = tdate.year
		xYr = tdate.year -1
		pbdate = [(cYr if cmmdd>x else xYr)*10000 + x for x in mmdd]
		df['pbdate'] = pbdate
		df['Portfolio'] = df['Portfolio'].apply(lambda x: x.strip())
		df['ticker'] = [s.replace('.','-') for s in df['Portfolio']]
		df = add_sector_industry(df)
	except Exception as e:
		sys.stderr.write('**ERROR: {} @ PULLING {}\n'.format(str(e),url))
		return {}
		
	if saveDB:
		mobj,_,msg = upsert_mdb(df,clientM=None,dbname=dbname,tablename=tablename,zpk={'ticker','pbdate'})
		sys.stderr.write('==Save to: MDB:{}: {}\n'.format(tablename,msg))

	return df

def opt_theme_list(argv,retParser=False):
	""" command-line options initial setup
	    Arguments:
		argv:	list arguments, usually passed from sys.argv
		retParser:	OptionParser class return flag, default to False
	    Return: (options, args) tuple if retParser is False else OptionParser class 
	"""
	parser = OptionParser(usage="usage: %prog [option] METHOD", version="%prog 0.7",
		description="get sp500_component and save to MDB::sp500_component")
	parser.add_option("","--no_database_save",action="store_false",dest="saveDB",default=True,
		help="no save to database (default: save)")
	parser.add_option("","--extra_js",action="store",dest="extraJS",
		help="extra JSON in DICT format.")
	parser.add_option("","--extra_xs",action="store",dest="extraXS",
		help="extra excutable string in k1=v1;k2=v2; format")
	parser.add_option("","--debug",action="store_true",dest="debugTF",default=False,
		help="debugging (default: False)")
	(options, args) = parser.parse_args(argv[1:])
	if retParser is True:
		return parser
	opts = vars(options)
	try:
		from _alan_str import extra_opts
		extra_opts(opts,xkey='extraJS',method='JS',updTF=True)
		extra_opts(opts,xkey='extraXS',method='XS',updTF=True)
	except Exception as e:
		sys.stderr.write("**ERROR: {}\n".format(str(e)))
	return (opts, args)

##### MAIN PROGRAM #####
if __name__ == '__main__':
	opts, args = opt_theme_list(sys.argv,retParser=False)
	if len(args)>0 and args[0] in globals():
		funcName = args[0]
	else:
		funcName = 'sp500_component'

	if funcName in globals():
		funcArg=globals()[funcName]
		sys.stderr.write("==RUN: {}\n".format(funcArg))
		df = funcArg(**opts)
	else:
		exit(-1)
	if isinstance(df,pd.DataFrame):
		if len(df)<30:
			sys.stderr.write("{}\n".format(df.to_csv(index=False,sep="|")))
		else:
			sys.stderr.write("{}\n".format(df.head(2).to_csv(index=False,sep="|")))
			sys.stderr.write("{}\n".format(df.tail(2).to_csv(index=False,sep="|")))
			sys.stderr.write("Total rows:{}\n".format(len(df)))
